/*
* BoardController.cpp
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/
//====================================================================================
// QT Includes
//====================================================================================
#include <QtCore/QDir>
#include <QtCore/QObject>
#include <QtCore/QPointF>
#include <QtGui/QPainter>
#include <QtGui/QImage>
#include <QtGui/QFileDialog>
#include <QtGui/QMessageBox>
#include <QtGui/QGraphicsView>
#include <QtGui/QGraphicsItem>
#include <QtXml/QXmlStreamWriter>
#include <QtXml/QXmlStreamReader>

//====================================================================================
// Project Includes
//====================================================================================
#include <iostream>
#include "log/draw.h"
#include "log/delete.h"
#include "log/modify.h"
#include "network/network.h"
#include "graphics/ContextRect.h"
#include "graphics/ContextErase.h"
#include "graphics/ContextEdit.h"
#include "graphics/ContextText.h"
#include "graphics/ContextEllipse.h"
#include "graphics/ContextSelect.h"
#include "graphics/ContextFreehand.h"
#include "graphics/ContextLine.h"
#include "graphics/ContextMove.h"
#include "icontext.h"
#include "idesc.h"
#include "idgraphicsitem.h"
#include "raknet/BitStream.h"

//====================================================================================
// Local Includes
//====================================================================================
#include "Caster.h"
#include "BoardController.h"

//------------------------------------------------------------------------------------
// Forward Declarations and Preprocessor Definitions
//------------------------------------------------------------------------------------


BoardController * BoardController::myself = 0;
using namespace RakNet;

BoardController::BoardController(QGraphicsScene* graphicsScene, QGraphicsView* graphicsView)
: sendTempEvents( true ), resumingSession(false)
{
  //set a reference to the scene and the view as part of MVC.
  boardModel = graphicsScene;
  boardView = graphicsView;
  //initialise the context to NULL.
  context = NULL;
  boardModel->setSceneRect(0, 0, 1280, 1024);
  Draw::RegisterCallback(this);
  Delete::RegisterCallback(this);
  Modify::RegisterCallback(this);
}

BoardController::~BoardController(void)
{
}

void BoardController::SquareButtonClicked()
{
  //change the square context.
  boardView->setCursor(Qt::CrossCursor);
  boardView->setDragMode(QGraphicsView::NoDrag);
  delete context;
  context = new ContextRect();
}

void BoardController::LineButtonClicked()
{
  //change the line context.
  boardView->setCursor(Qt::CrossCursor);
  boardView->setDragMode(QGraphicsView::NoDrag);
  delete context;
  context = new ContextLine();
}
void BoardController::TextButtonClicked()
{
  //change the text context.
  boardView->setCursor(Qt::IBeamCursor);
  boardView->setDragMode(QGraphicsView::NoDrag);
  delete context;
  context = new ContextText();
}
void BoardController::EllipseButtonClicked()
{
  //change the ellipse context.
  boardView->setCursor(Qt::CrossCursor);
  boardView->setDragMode(QGraphicsView::NoDrag);
  delete context;
  context = new ContextEllipse();
}
void BoardController::FreehandButtonClicked()
{
  //change the freehand context.
  boardView->setCursor(Qt::CrossCursor);
  boardView->setDragMode(QGraphicsView::NoDrag);
  delete context;
  context = new ContextFreehand();
}	

void BoardController::EraseButtonClicked()
{
  //change the erase context
  delete context;
  context = new ContextErase();
}

void BoardController::MoveButtonClicked()
{
  //change to the move context.
  boardView->setDragMode(QGraphicsView::NoDrag);
  boardView->setCursor(Qt::SizeAllCursor);
  delete context;
  context = new ContextMove();
}

void BoardController::SelectButtonClicked()
{
  //change to the selection context.
  boardView->setDragMode(QGraphicsView::RubberBandDrag);
  boardView->setCursor(Qt::CrossCursor);
  boardView->setMouseTracking(true);
  delete context;
  context = new ContextSelect();
}

void BoardController::PressCallBack(iDGraphicsItem* item)
{
  //get the correct context when superclass is called in the model
  if( context )
    context->MousePressCallBack(item);
}

void BoardController::MoveCallBack(iDGraphicsItem* item)
{
  //get the correct context when superclass is called in the model
  if( context )
    context->MouseMoveCallBack(item);
}

void BoardController::ReleaseCallBack(iDGraphicsItem* item)
{
  //get the correct context when superclass is called in the model
  if( context )
    context->MouseReleaseCallBack(item);
}

bool BoardController::DeleteAll()
{
  //delete all graphics on the board by itterating through.
  if(!resumingSession)
  {
    QList<QGraphicsItem *> list = this->boardModel->items();
    for (int i = 0; i < list.size(); i++){
      QGraphicsItem * targetItem = list.at(i);
      this->boardModel->removeItem(targetItem);
      delete targetItem;
    }
    return true;
  }

  resumingSession = false;
  return false;
}

void BoardController::AddEvent(const Event& e)
{
  // Ugly, but needed as we want to set the graphics item in GetFactory().
  Draw& draw = const_cast<Draw&>(dynamic_cast<const Draw&>(e));
  iDesc* desc = draw.GetDesc();
  // if an event has a description...
  if(desc)
  {
    //get a factory for this description.
    iDGraphicsItem* ditem = desc->GetFactory(this, &draw);
    //initialise the object (set focus etc...)
    ditem->OnItemSetup();

    QGraphicsItem* qitem = Caster::Cast(ditem);
    QGraphicsItem* parent = boardModel->itemAt(qitem->x(), qitem->y());
    if(parent)
    {
      //set parenting (relative position)
      qitem->setParentItem(parent);
      qitem->setPos( qitem->pos() - parent->scenePos() );
    }
    else
    {
      //add to board
      AddItem(qitem);
    }
    //if the parent is in the current branch...
    if(Log::GetSingleton().GetCurrentEvent() == e.GetParentEvent())
    {
      //...show the item
      qitem->show();
    }
    else
    {
      //..otherwise hide
      qitem->hide();
    }
  }
}

void BoardController::ModifyEvent(const Event& event)
{
  //delete a graphics item from the board based on the event.
  const Modify& modify = dynamic_cast<const Modify&>(event);
  if(!modify.GetDraw())
    return;

  QGraphicsItem* qitem = modify.GetDraw()->GetGraphicsItem();
  if(!qitem)
    return;

  // Handle the modify.
  if(modify.GetDesc())
  {
    // The following cast is required to turn Q into D to access handle() methods
    iDGraphicsItem* item = Caster::Cast(qitem);
    item->Modify(modify.GetDesc());
  }
  else
  {
    qitem->moveBy(modify.GetMove().x(), modify.GetMove().y());
  }
}

void BoardController::DeleteEvent(const Event& event)
{
  //delete a graphics item from the board based on the event.
  const Delete& deleteEvent = dynamic_cast<const Delete&>(event);
  const Draw* draw = deleteEvent.GetDraw();
  QGraphicsItem *item = draw->GetGraphicsItem();
  if( !item )
    return;

  if(!draw->IsToBeLogged())
  {
    RemoveItem(item);
  }
  else
  {
    item->hide();
  }
}



void BoardController::MoveObjects(QPointF& change)
{
  //create the modify events for every selected item when items are moved.
  if(change.isNull())
    return;

  QList<QGraphicsItem*> selectedItems = boardModel->selectedItems();
  for (int i = 0; i < selectedItems.size(); i++)
  {
    iDGraphicsItem* item = Caster::Cast(selectedItems.at(i));
    if(!item->GetLock())
    {
      Modify* modify = new TempModify(change, item->GetEvent());
      selectedItems.at(i)->show();
      selectedItems.at(i)->moveBy(modify->GetMove().x(), modify->GetMove().y());
      if( sendTempEvents )
      {
        modify->SendEvent(Network::GetSingletonPtr(), MEDIUM_PRIORITY, 0);
      }
    }
  }
}

void BoardController::EraseObjects()
{
  //create the delete event for every selected item. (For erasing multiple graphics)
  QList<QGraphicsItem*> selectedItems = boardModel->selectedItems();
  for (int i = selectedItems.size()-1; i >= 0; --i)
  {
    iDGraphicsItem* item = Caster::Cast(selectedItems.at(i));
    if(!item->GetLock())
    {
      Delete* d = new Delete(item->GetEvent());
      Log::GetSingleton().Add(d);
      d->SendEvent(Network::GetSingletonPtr(), MEDIUM_PRIORITY, 0);
    }
  }
}

void BoardController::InstantiateModifiers()
{
  //create a modify event for every selected item.(For moving multiple graphics)
  QList<QGraphicsItem*> selectedItems = boardModel->selectedItems();
  for (int i = 0; i < selectedItems.size(); i++)
  {
    iDGraphicsItem* item = Caster::Cast(selectedItems.at(i));
    if(!item->GetLock())
    {
      Modify* modify = new Modify(item->GetDesc(), item->GetEvent());
      Log::GetSingleton().Add(modify);
      modify->SendEvent(Network::GetSingletonPtr(), MEDIUM_PRIORITY, 0);
    }
  }
}

void BoardController::ClearSelection()
{
  //clear the selection on the board by iterating through the selected items.
  QList<QGraphicsItem*> selectedItems = boardModel->selectedItems();
  for (int i=0; i < selectedItems.size(); i++)
  {
    selectedItems.at(i)->setSelected(false);
  }
}

void BoardController::OnExportClicked()
{
  QString fileName = QFileDialog::getSaveFileName(this->boardView, tr("Save File"),"", tr("Images (*.png)"));
  if (!fileName.isEmpty()){
    //assuming we have a filename, save the contents of the model.
    //we need a QImage to map to.
    QImage myImage(this->boardModel->sceneRect().toRect().size(),QImage::Format_RGB32);
    QPainter painter(&myImage);
    //set the background to white otherwise it appears black.
    painter.setBackground(QBrush(QColor(255,255,255)));
    this->boardView->setBackgroundBrush(QBrush(QColor(255,255,255)));
    this->boardModel->setBackgroundBrush(QBrush(QColor(255,255,255)));
    //render the image to the QImage.
    this->boardModel->render(&painter,this->boardModel->sceneRect().toRect(),myImage.rect());
    painter.end();
    //write to disk.
    myImage.save(fileName,"PNG",-1);
    std::cout << "Image exported";
  }
}

void BoardController::OnUndoClicked()
{
  //if the undo button is pressed, call the jumpfunction in the log on the current event.
  Event* currentEvent = Log::GetSingleton().GetCurrentEvent();
  if(currentEvent)
  {
    Log::GetSingleton().JumpTo(currentEvent->GetParentEvent());
  }
}
void BoardController::OnRedoClicked()
{
  //if the redo button is pressed, the log moves back one undone event
  Log::GetSingleton().JumpBack();
}

void BoardController::SaveSessionClicked()
{
  //excute the save dialog box and return the filename. This dialogbox is native.
  QString fileName = QFileDialog::getSaveFileName(this->boardView, tr("Save File"),"", tr("Extensible Markup Language (*.xml)"));
  if (!fileName.isEmpty()){
    QFile file(fileName);
    if (!file.open(QFile::WriteOnly | QFile::Text)) {
      //         QMessageBox::warning(this, tr("Save Session"),
      //                              tr("Cannot write file %1:\n%2.")
      //                              .arg(fileName)
      //                              .arg(file.errorString()));
      return;
    }

    QXmlStreamWriter xml( &file );
    xml.setAutoFormatting( true );
    xml.writeStartDocument();
    xml.writeDTD("<!DOCTYPE tabula>");

    xml.writeStartElement( "tabula" );
    xml.writeAttribute( "version", "1.0" );

    Log::GetSingleton().Serialise( xml );

    xml.writeEndDocument();
  }
}

void BoardController::ResumeSessionClicked()
{
  resumingSession = true;
  //execute the
  QString fileName = QFileDialog::getOpenFileName(this->boardView, tr("Open File"),"", tr("Extensible Markup Language (*.xml)"));
  if (!fileName.isEmpty()){
    QFile file(fileName);
    if (!file.open(QFile::ReadOnly | QFile::Text)) {
      //         QMessageBox::warning(this, tr("Save Session"),
      //                              tr("Cannot write file %1:\n%2.")
      //                              .arg(fileName)
      //                              .arg(file.errorString()));
      return;
    }
    //look for the tabula tag, and returning if the version is incorrect.
    QXmlStreamReader xml( &file );
    while (!xml.atEnd())
    {
      xml.readNext();

      if (xml.isStartElement())
      {
        if (xml.name() == "tabula" && xml.attributes().value("version") == "1.0")
          break;
        else
        {
          xml.raiseError(QObject::tr("The file is not a tabula version 1.0 file."));
          return;
        }
      }
    }
    //look for the log tag. since this is all we are interested in.
    while( !xml.atEnd() )
    {
      xml.readNext();

      if( xml.isEndElement() )
        break;

      if( xml.isStartElement() )
      {
        if( xml.name() == "log" )
          Log::GetSingleton().Deserialise( xml );
      }
    }
    //any other searches for tags alternative to log can be done here
  }
}

void BoardController::FastConnectionClicked()
{
  //set our network state to slow connection. This will cause TEMPS to be broadcast.
  sendTempEvents = true;
}

void BoardController::SlowConnectionClicked()
{
  //set our network state to slow connection. This will cause TEEMPS not to be broadcast.
  sendTempEvents = false;
}
