/*
* BoardModel.cpp
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/
//====================================================================================
// QT Includes
//====================================================================================
#include <QtCore/QPointF>
#include <QtGui/QGraphicsSceneMouseEvent>
#include <QtGui/QGraphicsRectItem>

//====================================================================================
// Project Includes
//====================================================================================
#include "log/draw.h"
#include "network/network.h"
#include "raknet/BitStream.h"
#include "idgraphicsitem.h"
#include "icontext.h"

//====================================================================================
// Local Includes
//====================================================================================
#include "BoardModel.h"
#include "BoardController.h"
#include "ItemMenu.h"
#include "Caster.h"

//------------------------------------------------------------------------------------
// Forward Declarations and Preprocessor Definitions
//------------------------------------------------------------------------------------


using namespace RakNet;

BoardModel::BoardModel(void)
{
}

BoardModel::~BoardModel(void)
{
}

void BoardModel::mousePressEvent(QGraphicsSceneMouseEvent * mouseEvent){
  /*this is an over-ride method for the QGraphicsScene.
  we are taking control of this method in order to provide our own implementation.
  We later call the superclass to forward the mouse event to the rest of the 
  items on the canvas that interact with the mouse event.
  the view forwards us the MouseEvent data, we want to instantiate temp events and execute them locally.
  we only broadcast if in fast connection mode.
  */
  //get the boardcontroller
  BoardController& boardController = BoardController::GetSingleton();

  //if this is a rightclick event, instantiate the right-click menu.
  if (mouseEvent->button() == Qt::RightButton)
  {
    //retrieve the item, and send a reference to the menu, this is needed
    //to set the locked properties checkbox.
    QGraphicsItem * item = itemAt(mouseEvent->scenePos());
    iDGraphicsItem * dItem = Caster::Cast(item);
    if (dItem){
      ItemMenu * itemMenu = new ItemMenu(dItem);
      QPoint screenPos = mouseEvent->screenPos();
      itemMenu->popUp(&screenPos);
    }
  }
  //if this is a leftclick event, ensure a context has been set before calling the down method
  //within the context.
  else if (mouseEvent->button() == Qt::LeftButton)
  {
    if (boardController.GetContext() != NULL)
    {
      //send a null event through, by reference to the mousepress event.
      Event* newEvent = NULL;
      if(boardController.GetContext()->MousePressEvent(mouseEvent, newEvent) && newEvent != NULL)
      {
        //we only send to the network, in fast connection
        if( newEvent->IsToBeLogged() || boardController.GetSendTempEvents() )
          newEvent->SendEvent(Network::GetSingletonPtr(), MEDIUM_PRIORITY, 0);
        //we always add events to the log.
        Log::GetSingleton().Add(newEvent, true);
      }
    }
    //call the superclass for our default implementation
    QGraphicsScene::mousePressEvent(mouseEvent);	
  }
}

void BoardModel::mouseMoveEvent(QGraphicsSceneMouseEvent *mouseEvent)
{
  /*this is an over-ride method for the QGraphicsScene.
  we are taking control of this method in order to provide our own implementation.
  We later call the superclass to forward the mouse event to the rest of the 
  items on the canvas that interact with the mouse event.
  the view forwards us the MouseEvent data, we want to instantiate temp events and execute them locally.
  we only broadcast if in fast connection mode.
  */
  //get the boardcontroller
  BoardController& boardController = BoardController::GetSingleton();
  //ensure a context has been set before calling the down method
  //within the context.
  if (boardController.GetContext() != 0)
  {
    //send a null event through, by reference to the mousepress event.
    Event* newEvent = NULL;
    QPointF newPos;
    //send this event to the mousemoveevent in the context.
    if( boardController.GetContext()->MouseMoveEvent(mouseEvent, newPos, newEvent) )
    {
      boardController.MoveObjects(newPos);
    }
    if(newEvent)
    {
      //we only send to the network, in fast connection
      if( newEvent->IsToBeLogged() || boardController.GetSendTempEvents() )
        newEvent->SendEvent(Network::GetSingletonPtr(), MEDIUM_PRIORITY, 0);
      //we always add events to the log.
      Log::GetSingleton().Add(newEvent);
    }
  }
  //call the superclass for our default implementation
  QGraphicsScene::mouseMoveEvent(mouseEvent);
}

void BoardModel::mouseReleaseEvent(QGraphicsSceneMouseEvent * mouseEvent)
{
  /*this is an over-ride method for the QGraphicsScene.
  we are taking control of this method in order to provide our own implementation.
  We later call the superclass to forward the mouse event to the rest of the 
  items on the canvas that interact with the mouse event.
  the view forwards us the MouseEvent data, we want to instantiate temp events and execute them locally.
  we only broadcast if in fast connection mode.
  */
  //get the boardcontroller
  BoardController& boardController = BoardController::GetSingleton();
  //check if this is the left mouse button up event
  if (mouseEvent->button() == Qt::LeftButton)
  {
    //ensure we did set a context...
    if (boardController.GetContext() != NULL)
    {
      //set event to null and pass by reference
      Event* newEvent = NULL;
      if(boardController.GetContext()->MouseReleaseEvent(mouseEvent, newEvent))
      {
        //send a final modify event. not a temp.
        boardController.InstantiateModifiers();
      }
      
      //if a new event has been created (passed by reference - e.g. Draw)..
      if(newEvent != NULL)
      {
        //send the event accross the network and add to the log.
        newEvent->SendEvent(Network::GetSingletonPtr(), MEDIUM_PRIORITY, 0);
        Log::GetSingleton().RemoveTemp(newEvent);
        Log::GetSingleton().Add(newEvent);
      }
    }
    //send to the superclass for items to handle individually if necessary.
    QGraphicsScene::mouseReleaseEvent(mouseEvent);	
  }
}
