/*
* tabula.cpp
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/

//====================================================================================
// QT Includes
//====================================================================================
#include <QtCore/QDir>
#include <QtCore/QSettings>
#include <QtGui/QApplication>
//====================================================================================
// Project Includes
//====================================================================================
#include "network/network.h"
#include "session/session.h"
#include "logger/logger.h"

//====================================================================================
// Local Includes
//====================================================================================
#include "gui.h"
#include "tabula.h"

//------------------------------------------------------------------------------------
// Forward Declarations and Preprocessor Definitions
//------------------------------------------------------------------------------------
using namespace RakNet;
Tabula* Tabula::myself = NULL;
QApplication* Tabula::app = NULL;
QSettings* Tabula::settings = NULL;
Logger* Tabula::logger = NULL;
QDir* Tabula::pathToExec = NULL;

Tabula::Tabula()
{  
  // Create settings object.
  settings = new QSettings( pathToExec->absoluteFilePath("tabulaConfig.ini"), QSettings::IniFormat );

  // Associate stderr with the logger.
  Logger::getSingleton().add( std::cerr, settings->value( "logger/stderr", 40 ).toInt() );
  // Open other logfiles.
  settings->beginGroup( "logger" );
  QStringList keys = settings->allKeys();
  foreach( QString key, keys )
  {
    if( key != "stderr" )
      Logger::getSingleton().add( pathToExec->absoluteFilePath( key ).toStdString(), settings->value( key ).toInt() );
  }
  settings->endGroup();

  DBG( 40, "Tabula created" );
}

Tabula::~Tabula()
{
  delete logger;
  delete settings;
}

int Tabula::Run(int argc, char** argv)
{
  // Init Qt.
  QApplication qapp(argc, argv);
  app = &qapp;

  setPathToExec( argv[0] );

  // Run the program.
  bool ret = GetSingleton().Start();

  // Cleanup.
  delete myself;
  return (ret ? 0 : 1);
}

void Tabula::setPathToExec( const char* progname )
{
  // Get the current working directory.
  pathToExec = new QDir( progname );
  pathToExec->cdUp();
}

bool Tabula::Start()
{
  GUI::Open();

  // Get the network and log.
  Network* network = Network::GetSingletonPtr();
  Log* log = Log::GetSingletonPtr();

  // Pass pointer to log to the network.
  network->SetLog(log);

  // Create a new session and attach a view to it.
  GUI *g = &GUI::GetSingleton();
  session = new Session();
  g->window.ui.sessionView->setModel(dynamic_cast<Session*>(session));

  network->SetSession(session);
  log->SetSession(session);

  // Pass control to QT.
  bool ret = app->exec();

  delete log;

  Network::Disconnect();

  GUI::Close();

  return ret;
}

int main(int argc, char** argv)
{
  // Run the program.
  return Tabula::Run(argc, argv);
}

#if defined(_WIN32) && defined(NDEBUG)
int __stdcall WinMain(HINSTANCE hApp, HINSTANCE prev, __nullterminated char cmd, int show)
{
  (void)hApp;
  (void)show;
  (void)prev;
  (void)cmd;
  return main (__argc, __argv);
}
#endif
