/*
* person.h
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Dan Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/

#ifndef __PERSON_H__
#define __PERSON_H__

//====================================================================================
// QT Includes
//====================================================================================
#include <QtXml/QXmlStreamWriter>

//====================================================================================
// Project Includes
//====================================================================================
#include "raknet/RakString.h"
#include "raknet/BitStream.h"
#include "logger/logger.h"

//====================================================================================
// Local Includes
//====================================================================================

//------------------------------------------------------------------------------------
// Forward Declarations and Preprocessor Definitions
//------------------------------------------------------------------------------------

class Person
{
public:
  Person(short id, const char* username, const SystemAddress& address, const short timeDifference, const short time)
    : id(id), username(username), address(address), timeDifference(timeDifference), time(time)
  {
  }

  Person(const Person& person)
    : id(person.GetID()), username(person.GetUsername()), address(person.GetAddress()), timeDifference(person.GetTimeDifference()), time(time)
  {
  }

  inline const short GetID() const { return id; }
  inline const char* GetUsername() const { return username; }
  inline const SystemAddress& GetAddress() const { return address; }
  inline const short GetTimeDifference() const { return timeDifference; }
  inline const short GetTime() const { return time; }

	inline void Serialise( RakNet::BitStream& bs ) const
	{
		bs.WriteCompressed( address.binaryAddress );
		bs.WriteCompressed( address.port );
	}

	inline void Serialise( QXmlStreamWriter& xml ) const
	{
		xml.writeStartElement( "person" );
		xml.writeAttribute( "username", username.C_String() );
		xml.writeAttribute( "address", QString::number( address.binaryAddress ) );
		xml.writeAttribute( "port", QString::number( address.port ) );
		xml.writeEndElement();
	}

  inline void AddConnected(short id) { connectedIDs.push_back(id); }
  inline void AddConnected(Person* p) { connected.push_back(p); }
  inline void AddConnected(bool reconnect) { reconnecting.push_back(reconnect); }
  inline std::vector<Person*>::const_iterator const GetConnected() { return connected.begin(); }
  inline size_t GetConnectedCount() const { return connected.size(); }

  inline void UpdateConnected(Person* p)
  {
    for(size_t i=0; i<connectedIDs.size(); i++)
    {
      if(connectedIDs[i] == p->GetID())
      {
        AddConnected(p);
        connectedIDs.erase(connectedIDs.begin() + i);
      }
    }

    for(size_t i=0; i<connected.size(); i++)
    {
      connected[i]->UpdateConnected(p);
    }
  }

  inline bool ReconnectTo(const Person* p)
  {
    size_t index = FindConnected(p);
    if(index != size_t(-1))
    {
      return reconnecting[index];
    }

    return false;
  }

  inline size_t FindConnected(const Person* p) const
  {
    for(size_t i=0; i<connected.size(); i++)
    {
      if(p == connected[i])
      {
        return i;
      }
    }

    return size_t(-1);
  }

  inline bool DeleteConnected(const Person* p)
  {
    size_t index = FindConnected(p);
    if(index != size_t(-1))
    {
      connected.erase(connected.begin() + index);
      if(reconnecting.size() > index)
        reconnecting.erase(reconnecting.begin() + index);
      return true;
    }
    
    return false;
  }

private:
  short id;
  const RakNet::RakString username;
  const SystemAddress address;
  const short timeDifference;
  const short time;
  std::vector<Person*> connected;
  std::vector<short> connectedIDs;
  std::vector<bool> reconnecting;
};

#endif // __PERSON_H__
