/*
* DGraphicsTextItem.cpp
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/

//====================================================================================
// QT Includes
//====================================================================================
#include <QtGui/QGraphicsSceneMouseEvent>
#include <QtCore/QString>
#include <QtGui/QTextDocument>
#include <QtGui/QGraphicsTextItem>
#include <QtCore/QPointF>
#include <QtGui/QTextCursor>
#include <QtGui/QPen>
#include <QtGui/QColor>
//====================================================================================
// Project Includes
//====================================================================================
#include <iostream>
#include "network/network.h"
#include "log/draw.h"
#include "log/modify.h"
//====================================================================================
// Local Includes
//====================================================================================
#include "DGraphicsTextItem.h"
#include "icontext.h"
#include "icontextcallback.h"
#include "ResizeDialog.h"
#include "DescText.h"
//------------------------------------------------------------------------------------
// Forward Declarations
//------------------------------------------------------------------------------------
typedef Modify ModifyEvent;

// Sets whether the item is locked for editing
void DGraphicsTextItem::SetLock(bool lock)
{
  locked = lock;
}

// Returns whether the object is locked
bool DGraphicsTextItem::GetLock()
{
  return locked;
}

// Brings an item forward a layer
void DGraphicsTextItem::BringForward()
{
}

// Pushes an item back a layer
void DGraphicsTextItem::SendBack()
{
}

// Constructor for text item
DGraphicsTextItem::DGraphicsTextItem(const QString &text, iContextCallback* callback) : callback(callback)
{
  QTextDocument *doc = new QTextDocument(text) ;
  setDocument(doc);
  setFlag(DGraphicsTextItem::ItemIsSelectable, true);
  setFlag(DGraphicsTextItem::ItemIgnoresTransformations, true);
  setFlag(DGraphicsTextItem::ItemIsFocusable, true);
  setZValue(0);
  locked = false;
}

// Defines behaviour for the mouse being pressed
void DGraphicsTextItem::mousePressEvent(QGraphicsSceneMouseEvent *mouseEvent)
{
  QGraphicsTextItem::mousePressEvent(mouseEvent);
  callback->PressCallBack(this);
}

// Defines behaviour for the text being moved
void DGraphicsTextItem::mouseMoveEvent(QGraphicsSceneMouseEvent *mouseEvent)
{
  QGraphicsTextItem::mouseMoveEvent(mouseEvent);
  callback->MoveCallBack(this);
}

// Defines behaviour for the mouse being released
void DGraphicsTextItem::mouseReleaseEvent(QGraphicsSceneMouseEvent *mouseEvent)
{
  QGraphicsTextItem::mouseReleaseEvent(mouseEvent);
  callback->ReleaseCallBack(this);
}

// Defines properties when the item is created, 
// i.e. that it is editable, where the cursor appears
void DGraphicsTextItem::OnItemSetup()
{
  setFocus();
  setTextInteractionFlags(Qt::TextEditable);
  QTextCursor tc = textCursor();
  tc.movePosition(QTextCursor::EndOfLine);
  setTextCursor(tc);
}

// Defines behaviour when mouse is double clicked on item.
void DGraphicsTextItem::mouseDoubleClickEvent(QGraphicsSceneMouseEvent *mouseEvent)
{
  setTextInteractionFlags(Qt::TextEditable);
  QTextCursor tc = textCursor();
  tc.movePosition(QTextCursor::EndOfLine);
  setTextCursor(tc);
}

// Defines what to do when focus moves somewhere away from the item
void DGraphicsTextItem::focusOutEvent(QFocusEvent *focusEvent)
{
  if(draw->IsToBeLogged())
  {
    setTextInteractionFlags(Qt::NoTextInteraction);
    ModifyEvent* modify = new ModifyEvent(GetDesc(), draw);
    Log::GetSingleton().Add(modify);
    modify->SendEvent(Network::GetSingletonPtr(), MEDIUM_PRIORITY, 0);
  }
}

// Destructor for item
DGraphicsTextItem::~DGraphicsTextItem(void)
{
}

// Defines what happens when the item is modified in terms of
// making sure it is locked so other users can't delete or
// move it while it is beign edited.
// Also defines new position if moved, or changes text if edited.
void DGraphicsTextItem::Modify(const iDesc* desc)
{
  const DescText* descT = dynamic_cast<const DescText*>(desc);
  if(descT)
  {
    locked = descT->GetLocked();

    if( parentItem() )
      setPos( parentItem()->mapFromScene( QPointF( descT->GetX(), descT->GetY() ) ) );
    else
      setPos(descT->GetX(), descT->GetY());
    setTextWidth(descT->GetWidth());
    setZValue(descT->GetZOrder());

    QTextDocument *doc = new QTextDocument(descT->GetText()) ;
    setDocument(doc);
  }
}

// Returns a description of the item, in this case, the plaintext of its contents,
// its position and whether it is locked
iDesc* DGraphicsTextItem::GetDesc()
{
  return static_cast<iDesc*>(new DescText(toPlainText(), scenePos().x(), scenePos().y(), textWidth(), zValue(), locked));
}

// Defines how to erase the item from the board
void DGraphicsTextItem::Erase()
{
  QGraphicsItem *item = this;
  item->hide();
}

void DGraphicsTextItem::SetEvent(Draw* d)
{
  draw = d;
  draw->SetGraphicsItem(this);
}

void DGraphicsTextItem::hoverEnterEvent(QGraphicsSceneHoverEvent *sceneEvent)
{
  //	QAbstractGraphicsShapeItem *item = this;
  //save the current colour
  //set the new pen of your choice;
  //item->setPen(QPen(QColor(255, 0, 0)));
}
void DGraphicsTextItem::hoverLeaveEvent(QGraphicsSceneHoverEvent *sceneEvent)
{
  //restore the original colour
  //	QAbstractGraphicsShapeItem *item = this;
  //save the current colour
  //set the new pen of your choice;
  //item->setPen(QPen(QColor(0, 0, 0)));
}

// Defines what to do on the start of a resize
void DGraphicsTextItem::ResizeStart( bool sendTempEvents )
{
  //ResizeDialog * resizeDialog = new ResizeDialog(this,this->rect().width(),1280,1024);
}

// Adjusts horizontal size as slider is moved
void DGraphicsTextItem::AdjustHorizontal(int value)
{
}

// Adjusts vertical size as slider is moved
void DGraphicsTextItem::AdjustVertical(int value)
{
}

// Sets final size when resize box is closed
void DGraphicsTextItem::ResizeEnd(QDialog *dialog, int x, int y)
{
}
