/*
* DescPath.cpp
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/
//====================================================================================
// QT Includes
//====================================================================================
#include <QtXml/QXmlStreamReader>
//====================================================================================
// Project Includes
//====================================================================================
#include "raknet/BitStream.h"
#include "log/draw.h"

//====================================================================================
// Local Includes
//====================================================================================
#include "DescPath.h"
#include "DGraphicsPathItem.h"

//------------------------------------------------------------------------------------
// Forward Declarations and Preprocessor Definitions
//------------------------------------------------------------------------------------
using namespace RakNet;

DescPath::DescPath(std::vector<QPointF> points, short zOrder, bool locked)
: points(points), factory(0), zOrder(zOrder), locked(locked)
{
}

DescPath::DescPath(BitStream& bs) : factory(0)
{
  // Read locals from BitStream
  bs.ReadCompressed(zOrder);
  bs.ReadCompressed(locked);
  unsigned int size;
  bs.ReadCompressed(size);

  // Read all points in path from BitStream
  for(unsigned int i=0; i<size; i++)
  {
    short x;
    short y;
    bs.ReadCompressed(x);
    bs.ReadCompressed(y);
    points.push_back(QPointF(x, y));
  }
}

DescPath::DescPath( QXmlStreamReader& xml ) : factory(0)
{
  // Get number of points & allocate
  size_t size = xml.attributes().value( "size" ).toString().toUInt();
  points.reserve( size );

  // Parse all points
  while( !xml.atEnd() )
  {
    xml.readNext();

    if( xml.isEndElement() )
      break;

    if( xml.isStartElement() )
    {
      // Confirm point and extract from XML to local
      Q_ASSERT( xml.name() == "point" );
      qreal x = xml.attributes().value("x").toString().toDouble();
      qreal y = xml.attributes().value("y").toString().toDouble();
      points.push_back( QPointF( x, y ) );

      // Loop till end of XML
      while( !xml.atEnd() && !xml.isEndElement() )
        xml.readNext();
    }
  }
}

DescPath::~DescPath(void)
{
}

void DescPath::Serialise(RakNet::BitStream &bs) const
{
  // Write path locals to BitStream
  bs.WriteCompressed((iDesc::DescType)Freehand);
  bs.WriteCompressed(zOrder);
  bs.WriteCompressed(locked);
  bs.WriteCompressed((unsigned int)points.size());

  // Write all points in path to XML
  for(unsigned int i=0; i<points.size(); i++)
  {
    short x = points[i].x();
    short y = points[i].y();
    bs.WriteCompressed(x);
    bs.WriteCompressed(y);
  }
}

void DescPath::Serialise(QXmlStreamWriter& xml) const
{
  // Write type to XML
  xml.writeStartElement( "desc" );
  xml.writeAttribute( "type", "freehand" );

  // Write how many points
  xml.writeAttribute( "size", QString::number( (uint)points.size() ) );
  
  // Write all points from locals to XML
  for(size_t i=0; i<points.size(); i++)
  {
    xml.writeStartElement( "point" );
    xml.writeAttribute( "x", QString::number( points[i].x() ) );
    xml.writeAttribute( "y", QString::number( points[i].y() ) );
    xml.writeEndElement();
  }

  xml.writeEndElement();
}

iDGraphicsItem* DescPath::GetFactory(iContextCallback* callback, Draw* draw)
{
  if(!factory)
  {
    // Create new path
    factory = new DGraphicsPathItem(callback);
    factory->SetEvent(draw);
    factory->setZValue(zOrder);
  }

  // Get points
  const std::vector<QPointF>& nPoints = factory->GetPoints();
  for(size_t i=nPoints.size(); i<points.size(); i++)
  {
    if(i==0)
    {
      // Write first
      QPainterPath p;
      p.moveTo(points[0]);
      factory->AddPoint(points[0]);
      factory->setPath(p);
    }
    else
    {
      // Add all points
      QPainterPath p = factory->path();
      p.lineTo(points[i]);
      factory->AddPoint(points[i]);
      factory->setPath(p);
    }
  }

  factory->setSelected(false);
  return factory;
}
