/*
* chat.cpp
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/

//====================================================================================
// QT Includes
//====================================================================================
#include <QtCore/QTime>
#include <QtCore/QString>
#include <QtCore/QByteArray>
#include <QtXml/QXmlStreamWriter>
//====================================================================================
// Project Includes
//====================================================================================
#include "network/messageids.h"
#include "person.h"
#include "raknet/BitStream.h"
#include "raknet/StringCompressor.h"
#include "logger/logger.h"
//====================================================================================
// Local Includes
//====================================================================================
#include "ilogcallback.h"
#include "chat.h"
//------------------------------------------------------------------------------------
// Forward Declarations
//------------------------------------------------------------------------------------

iLogCallback* Chat::callback = NULL;


Chat::Chat( const QString& chatText, const QTime& chatTime, const Person* sender )
: Event( true, sender, false ), text( chatText), time( chatTime ), sender( sender->GetUsername() )
{
}

Chat::Chat( RakNet::BitStream& bs, const iSession* session )
: Event( bs, session ), sender( owner->GetUsername() ) // deserialise recursively
{
  size_t length;
  bs.Read( length );
  char* buf = new char[length];
  stringCompressor->DecodeString(buf, (int)length, &bs);
  text = QString( buf );
  delete buf;

  int h, m, s;

  bs.ReadCompressed(h);
  bs.ReadCompressed(m);
  bs.ReadCompressed(s);
  time = QTime( h, m, s );
}

Chat::~Chat()
{}

QString Chat::GetSender() const
{
  return sender;
}

QString Chat::GetText() const
{
  return text;
}

QTime Chat::GetTime() const
{
  return time;
}

void Chat::Execute() const
{
  if( callback )
    callback->AddEvent(*this);
}

void Chat::SendEvent(iNetwork* network, PacketPriority priority, SystemAddress* address)
{
  RakNet::BitStream bs;
  Serialise(bs);
  network->SendBitStream(bs, priority, RELIABLE_ORDERED, 1, address);
}

void Chat::Serialise( RakNet::BitStream& bs ) const
{
  bs.Write((MessageID)ID_CHAT_EVENT);
  bs.Write((unsigned char)RELIABLE_ORDERED);
  bs.Write((char)1);

  Event::Serialise( bs );

  QByteArray buf = text.toUtf8();
  bs.Write( (size_t) ( buf.length()+1 ) );
  stringCompressor->EncodeString( buf, buf.length()+1, &bs );

  bs.WriteCompressed(time.hour());
  bs.WriteCompressed(time.minute());
  bs.WriteCompressed(time.second());
}

void Chat::Serialise( QXmlStreamWriter& xml ) const
{
  xml.writeStartElement( "chat" );
  xml.writeAttribute( "time", time.toString() );
  xml.writeAttribute( "sender", sender );

  xml.writeStartElement( "text" );
  xml.writeCharacters( text );
  xml.writeEndElement();

  Event::Serialise( xml );

  xml.writeEndElement();
}

Chat::Chat( QXmlStreamReader& xml, const iSession* session )
{
  Q_ASSERT( xml.isStartElement() && xml.name() == "chat" );
  time = QTime::fromString( xml.attributes().value( "time" ).toString() );
  sender = xml.attributes().value( "sender" ).toString();

	// loop over nested XML elements
  while( !xml.atEnd() )
  {
    xml.readNext();

    if( xml.isEndElement() )
      break;

    if( xml.isStartElement() )
    {
			// reading the inherited event
      if( xml.name() == "event" )
      {
        Event::Deserialise( xml, session );
      }
			// reading the text message
      else if( xml.name() == "text" )
      {
        xml.readNext();
        text = "";
        while( !xml.isEndElement() )
        {
          text += xml.text().toString();
          xml.readNext();
        }
      }
    }
  }
}

