/*
* delete.h
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/
#ifndef __DELETE_H__
#define __DELETE_H__

//====================================================================================
// QT Includes
//====================================================================================

//====================================================================================
// Project Includes
//====================================================================================

//====================================================================================
// Local Includes
//====================================================================================
#include "event.h"

//------------------------------------------------------------------------------------
// Forward Declarations
//------------------------------------------------------------------------------------
class Draw;
namespace RakNet
{
  class BitStream;
}
class QXmlStreamWriter;
class QXmlStreamReader;

struct iLogCallback;
struct iSession;

class Delete: public Event
{
public:
	/**
	 * Registers the callback which will be called on Execute()
	 * @arguments cbObject Callback to be registered.
	 */
  static void RegisterCallback( iLogCallback* cbObject )
  { callback = cbObject; }

	/**
	 * Create an event which will cause the passed draw event to disappear from the screen.
	 * @argument drawEvent The shape on the screen to be hidden.
	 * @argument isToBeLogged If set to true this event would not be stored in the log though it is added.
	 */
  Delete( Draw* drawEvent, bool isToBeLogged = true );

	/**
	 * Deserialise a delete event from a network bitstream.
	 * @argument bs Bitstream to read from.
	 * @argument session Current session to deserialise the owner of this event.
	 */
  Delete( RakNet::BitStream& bs, const iSession* session );

	/**
	 * Deserialise a delete event from a Qt XML stream.
	 * @argument bs Qt XML stream to read from. The current token should be a chat tag.
	 * @argument session Current session to deserialise the owner of this event.
	 */
  Delete( QXmlStreamReader& xml, const iSession* session );

  virtual ~Delete();

  const Draw* GetDraw() const;

	/// If a callback is registered (@see RegisterCallback), it will be called with this event as an argument.
  void Execute() const;

 	/**
	 * Send this event to the network object.
	 * @argument network Network instance to send to.
	 * @argument priority Send priority for the network. High prioritised messages might overtake lower prioritised in the local buffer.
	 * @argument address The raknet address of the destination. If NULL it will be broadcasted.
	 */
 virtual void SendEvent(iNetwork* network, PacketPriority priority, SystemAddress* address);

	/**
	 * Serialise this event to a Qt XML stream.
	 * @argument xml XML stream to write to.
	 */
  virtual void Serialise( QXmlStreamWriter& xml ) const;

	/// Shows the referenced draw event again.
  virtual void Undo();

	/// Hides the referenced event again.
  virtual void Redo();

	/// Returns a description to be shown in the history view.
  virtual QString GetStringDesc() const;

protected:
private:
  static iLogCallback* callback;

  virtual void Serialise( RakNet::BitStream& bs ) const;

  Draw* draw;
};

class TempDelete : public Delete
{
public:
	/**
	 * Create a temporary delete event. It behaves like a delete event but will not be stored in the log.
	 * @argument drawEvent Event to be hidden.
	 */
  TempDelete(Draw* drawEvent);
  virtual ~TempDelete() {}
};

#endif // __DELETE_H__
