/*
* draw.cpp
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/

//====================================================================================
// QT Includes
//====================================================================================
#include <QtGui/QGraphicsItem>
#include <QtXml/QXmlStreamWriter>
#include <QtXml/QXmlStreamReader>
//====================================================================================
// Project Includes
//====================================================================================
#include "idesc.h"
#include "network/messageids.h"
#include "raknet/BitStream.h"
#include "session/session.h"
#include "graphics/DescEllipse.h"
#include "graphics/DescLine.h"
#include "graphics/DescRect.h"
#include "graphics/DescText.h"
#include "graphics/DescPath.h"
#include "graphics/DescDeserialiser.h"
//====================================================================================
// Local Includes
//====================================================================================
#include "ilogcallback.h"
#include "draw.h"
//------------------------------------------------------------------------------------
// Forward Declarations
//------------------------------------------------------------------------------------
iLogCallback* Draw::callback = NULL;

Draw::Draw(iDesc* desc, bool isToBeLogged)
: Event(isToBeLogged), desc(desc)
{
}

TempDraw::TempDraw(iDesc* desc)
: Draw(desc, false)
{
}

Draw::Draw(RakNet::BitStream& bs, const iSession* session)
: Event(bs, session), qItem(0) // deserialise recursively
{
	// find out which type descripter is to be used
  iDesc::DescType type;
  bs.ReadCompressed(type);

  switch(type)
  {
  case iDesc::Ellipse:
    {
      desc = new DescEllipse(bs);
      break;
    }
  case iDesc::Line:
    {
      desc = new DescLine(bs);
      break;
    }
  case iDesc::Rect:
    {
      desc = new DescRect(bs);
      break;
    }
  case iDesc::Text:
    {
      desc = new DescText(bs);
      break;
    }
  case iDesc::Freehand:
    {
      desc = new DescPath(bs);
      break;
    }
  }

}

void Draw::Execute() const
{
  if(callback)
  {
    callback->AddEvent(*this);
  }
}

void Draw::SendEvent(iNetwork* network, PacketPriority priority, SystemAddress* address)
{
  RakNet::BitStream bs;
  Serialise(bs);
  network->SendBitStream(bs, priority, RELIABLE_ORDERED, 2, address);
}

void Draw::Serialise(RakNet::BitStream& bs) const
{
  bs.Write((MessageID)ID_DRAW_EVENT);
  bs.Write((unsigned char)RELIABLE_ORDERED);
  bs.Write((char)2);

  Event::Serialise(bs);

  desc->Serialise(bs);
}

void Draw::Serialise(QXmlStreamWriter& xml) const
{
  xml.writeStartElement( "draw" );

  desc->Serialise(xml);

	// serialise the superclass as well
  Event::Serialise(xml);

  xml.writeEndElement();
}

Draw::Draw( QXmlStreamReader& xml, const iSession* session )
{
  Q_ASSERT( xml.isStartElement() && xml.name() == "draw" );

	// parse nested tags
  while( !xml.atEnd() )
  {
    xml.readNext();

    if( xml.isEndElement() )
      break;

    if( xml.isStartElement() )
    {
			// parse descriptor
      if( xml.name() == "desc" )
        desc = DescDeserialiser::Deserialise( xml );
			// parse inherited event
      else if( xml.name() == "event" )
        Event::Deserialise( xml, session );
    }
  }
}

void Draw::Undo()
{
  qItem->hide();
}

void Draw::Redo()
{
  qItem->show();
}

QString Draw::GetStringDesc() const
{
  return QString( "draw " ) + desc->GetStringDesc();
}

