/*
* draw.h
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/
#ifndef __DRAW_H__
#define __DRAW_H__

//====================================================================================
// QT Includes
//====================================================================================

//====================================================================================
// Project Includes
//====================================================================================

//====================================================================================
// Local Includes
//====================================================================================
#include "event.h"

//------------------------------------------------------------------------------------
// Forward Declarations
//------------------------------------------------------------------------------------
class QGraphicsItem;
class QXmlStreamWriter;
class QXmlStreamReader;
struct iLogCallback;
struct iDesc;
struct iSession;
namespace RakNet
{
  class BitStream;
}

class Draw: public Event
{
public:
	/**
	 * Registers the callback which will be called on Execute()
	 * @arguments cbObject Callback to be registered.
	 */
  static void RegisterCallback( iLogCallback* cbObject )
  { callback = cbObject; }

 	/**
	 * Cause an shape to be created on the canvas.
	 * @argument desc Descriptor object of the shape.
	 * @argument isToBeLogged If set to true this event would not be stored in the log though it is added.
	 */
	Draw(iDesc* desc, bool isToBeLogged = true);

	/**
	 * Deserialise a draw event from a network bitstream.
	 * @argument bs Bitstream to read from.
	 * @argument session Current session to deserialise the owner of this event.
	 */
  Draw(RakNet::BitStream& bs, const iSession* session);

	/**
	 * Deserialise a draw event from a Qt XML stream.
	 * @argument bs Qt XML stream to read from. The current token should be a chat tag.
	 * @argument session Current session to deserialise the owner of this event.
	 */
  Draw( QXmlStreamReader& xml, const iSession* session );

  virtual ~Draw() {}

	/**
	 * Store the actual Qt item on the screen so modify events pointing
	 * to this event can access it.
	 * @argument item The Qt item on the screen.
	 */
  void SetGraphicsItem(QGraphicsItem* item) { qItem = item; }
  QGraphicsItem* GetGraphicsItem() const { return qItem; }

	/// If a callback is registered (@see RegisterCallback), it will be called with this event as an argument.
  void Execute() const;

  inline iDesc* GetDesc() const { return desc; }

 	/**
	 * Send this event to the network object.
	 * @argument network Network instance to send to.
	 * @argument priority Send priority for the network. High prioritised messages might overtake lower prioritised in the local buffer.
	 * @argument address The raknet address of the destination. If NULL it will be broadcasted.
	 */
  virtual void SendEvent(iNetwork* network, PacketPriority priority, SystemAddress* address);

	/**
	 * Serialise this event to a Qt XML stream.
	 * @argument xml XML stream to write to.
	 */
  virtual void Serialise( QXmlStreamWriter& xml ) const;

	/// Hide the graphics item on the screen again.
  virtual void Undo();

	/// Show the graphics item again.
  virtual void Redo();

	/// Returns a description to be shown in the history view.
  virtual QString GetStringDesc() const;

private:
  static iLogCallback* callback;

  virtual void Serialise( RakNet::BitStream& bs ) const;

  iDesc* desc;
  QGraphicsItem* qItem;
};

class TempDraw : public Draw
{
public:
	/**
	 * Create a temporary draw event. It behaves like a draw event but will not be stored in the log.
	 * @argument desc Descriptor of the shape.
	 */
  TempDraw(iDesc* desc);
  virtual ~TempDraw() {}
};

#endif


