/*
* event.cpp
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/

//====================================================================================
// QT Includes
//====================================================================================
#include <QtXml/QXmlStreamWriter>
#include <QtXml/QXmlStreamReader>
//====================================================================================
// Project Includes
//====================================================================================
#include "raknet/BitStream.h"
#include "isession.h"
#include "person.h"
#include "logger/logger.h"
//====================================================================================
// Local Includes
//====================================================================================
#include "log.h"
#include "timestamp.h"
#include "event.h"
//------------------------------------------------------------------------------------
// Forward Declarations
//------------------------------------------------------------------------------------


Event::Event( bool isToBeLogged, const Person* eventOwner, bool parent )
: parentEvent(parent ? Log::GetSingleton().GetCurrentEvent() : 0), isToBeLogged(isToBeLogged),
owner(eventOwner), operationRequired(false)
{
  if( isToBeLogged )
  {
    virtualTime = ++Log::Now();
  }
  else
  {
    virtualTime = Timestamp();
  }

  if(parentEvent)
  {
    if(isToBeLogged)
    {
      parentEvent->AddChildEvent(this);
    }
    level = parentEvent->GetLevel()+1;
  }
  else
  {
    level = 1;
  }
}

Event::Event( RakNet::BitStream& bs, const iSession* session)
: parentEvent(0), owner(session->Deserialize(bs)), operationRequired(false)
{
  bs.ReadCompressed(isToBeLogged);

  bool firstEvent;
  bs.ReadCompressed(firstEvent);

  if(!firstEvent)
  {
    Timestamp parent(bs);
    parentEvent = Log::GetSingleton().FindEvent(parent);
    if(isToBeLogged)
    {
      parentEvent->AddChildEvent(this);
    }
    level = parentEvent->GetLevel()+1;
  }
  else
  {
    level = 1;
  }

  if(isToBeLogged)
  {
    virtualTime = Timestamp(bs);
  }
  else
  {
    virtualTime = Timestamp();
  }
}

Event::Event(const Timestamp& ts) : parentEvent(0), isToBeLogged(true), owner(NULL), operationRequired(false), virtualTime(ts)
{}

Event::Event() : parentEvent(0), isToBeLogged(true), owner(NULL), operationRequired(false)
{}

Event::~Event()
{
}

const Timestamp& Event::GetTimestamp() const
{
  return virtualTime;
}

const Person* Event::GetOwner() const
{
  return owner;
}

bool Event::operator<( const Event& e ) const
{
  return this->virtualTime < e.virtualTime;
}

bool Event::operator>( const Event& e ) const
{
  return this->virtualTime > e.virtualTime;
}

bool Event::operator==( const Event& e ) const
{
  return this->virtualTime == e.virtualTime;
}

bool Event::operator||( const Event& e ) const
{
  return this->virtualTime || e.virtualTime;
}

void Event::Serialise( RakNet::BitStream& bs ) const
{
  owner->Serialise( bs );
  bs.WriteCompressed(isToBeLogged);
  if(parentEvent)
  {
    bs.WriteCompressed(false);
    parentEvent->GetTimestamp().Serialise( bs );
  }
  else
  {
    bs.WriteCompressed(true);
  }
  if(isToBeLogged)
  {
    virtualTime.Serialise( bs );
  }
}

void Event::Serialise( QXmlStreamWriter& xml ) const
{
  xml.writeStartElement( "event" );
  //	xml.writeAttribute( "owner", owner->GetUsername() );
  virtualTime.Serialise( xml );

  if( parentEvent )
  {
    xml.writeStartElement( "parent" );
    parentEvent->GetTimestamp().Serialise( xml );
    xml.writeEndElement();
  }

  xml.writeEndElement();
}

void Event::Deserialise( QXmlStreamReader& xml, const iSession* session )
{
  Q_ASSERT( xml.isStartElement() && xml.name() == "event" );
  owner = Log::GetSingleton().GetOwner();
  //	owner = session->FindMember( xml.attributes().value("owner").toString().toUtf8().constData() );

  parentEvent = NULL;
  while( !xml.atEnd() )
  {
    xml.readNext();

    if( xml.isEndElement() )
      break;

    if( xml.isStartElement() )
    {
      if( xml.name() == "timestamp" )
      {
        virtualTime = Timestamp( xml );
      }
      else if( xml.name() == "parent" )
      {
        while( !xml.atEnd() )
        {
          xml.readNext();

          if( xml.isEndElement() )
            break;

          if( xml.isStartElement() && xml.name() == "timestamp" )
          {
            Timestamp parent( xml  );
            parentEvent = Log::GetSingleton().FindEvent( parent );
          }
        }
      }
    }
  }

  if( parentEvent )
  {
    parentEvent->AddChildEvent(this);
    level = parentEvent->GetLevel() + 1;
  }
  else
    level = 1;
}



int Event::row() const
{
  if (parentEvent)
    return parentEvent->childEvents.indexOf(const_cast<Event*>(this));
  return 0;
}
