/*
* event.h
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/
#ifndef __EVENT_H__
#define __EVENT_H__

//====================================================================================
// QT Includes
//====================================================================================

//====================================================================================
// Project Includes
//====================================================================================
#include <QtCore/QList>
#include <list>
#include "inetwork.h"
#include "raknet/PacketPriority.h"

//====================================================================================
// Local Includes
//====================================================================================
#include "timestamp.h"
#include "log.h"

//------------------------------------------------------------------------------------
// Forward Declarations
//------------------------------------------------------------------------------------
class Person;
struct iSession;
namespace RakNet
{
  class BitStream;
}
class QXmlStreamWriter;
class QXmlStreamReader;

/**
* Event is everything happening in a conversation worth keeping it in the Log.
* This is an interface everything which wants to become logged should inherit from.
*/
class Event
{
public:
  /**
  * Create a new Event.
	* @argument isToBeLogged If set to true this event would not be stored in the log though it is added.
  * @argument eventOwner Specifies the owner of the event. The owner might have special
  *                      permissions to modify it.
	* @argument parent Specifies if this event should look up the current event in the log to add it as a parent.
  */
  Event( bool isToBeLogged, const Person* eventOwner = Log::GetSingleton().GetOwner(), bool parent = true );

	/**
	 * Deserialise an event from a network bitstream.
	 * @argument bs Bitstream to read from.
	 * @argument session Current session to deserialise the owner of this event.
	 */
  Event( RakNet::BitStream& bs, const iSession* session );

	/**
	 * Deserialise a draw event from a Qt XML stream.
	 * @argument bs Qt XML stream to read from. The current token should be a chat tag.
	 * @argument session Current session to deserialise the owner of this event.
	 */
  Event( const Timestamp& ts );


	/// Create a blank useless event. Use Deserialise() afterwards.
  Event();

  /// Virtual destructor.
  virtual ~Event();

	/// Just an interface.
  virtual void Execute() const { }

  /**
  * Selector for the logic time at which this event took place.
  * @returns The time at which the event took place.
  */
  const Timestamp& GetTimestamp() const;

  /**
  * Selector for the author of this event.
  * @returns The auther of this event.
  */
  const Person* GetOwner() const;

  /**
  * Checks whether this event happened earlier than e.
  */
  bool operator<( const Event& e ) const;

  /**
  * Checks whether this event happened later than e.
  */
  bool operator>( const Event& e ) const;

  /**
  * Checks whether this event is the same as e.
  */
  bool operator==( const Event& e ) const;

  /**
  * Checks whether this event happened concurrent to e.
  */
  bool operator||( const Event& e ) const;

	/**
	 * Serialise this event to a Qt XML stream.
	 * @argument xml XML stream to write to.
	 */
  virtual void Serialise( QXmlStreamWriter& xml ) const;

	/// Return if this event should be stored in the log or not.
  bool IsToBeLogged() const { return isToBeLogged; }

	/**
	 * When some events adds this one as a parent it has to add itself as a child here.
	 * @argument e The child event to be added.
	 */
  void AddChildEvent(Event* e) { childEvents.append(e); e->SetParentEvent(this); }

	/// Returns all registered children events.
  QList<Event*>::const_iterator GetChildren() const { return childEvents.begin(); }

  int GetChildrenCount() const { 
    return childEvents.size(); 
  }

	/// Remove children, do not deallocate them.
  void ClearChildren()
  {
    childEvents.clear();
  }

	/**
	 * Return child number row.
	 * @argument row The index of the child to be returned.
	 */
  Event* child(int row)
  {
    return this->childEvents.value(row);
  }

  void SetParentEvent(Event* e) { parentEvent = e; }

  int columnCount() const { return 1; }

	/// Returns the child index of this event in the parents child list.
  int row() const;

  Event* GetParentEvent() const { return parentEvent; }

  /// Undo / Redo interface.
  virtual void Undo() { }
  virtual void Redo() { }

  size_t GetLevel() const { return level; }
  bool IsOperationRequired() const { return operationRequired; }
  void MarkOperationRequired() { operationRequired = true; }
  void MarkOperationNotRequired() { operationRequired = false; }

 	/**
	 * Serialise this event to a Qt XML stream.
	 * @argument xml XML stream to write to.
	 */
	virtual void SendEvent(iNetwork* network, PacketPriority priority, SystemAddress* address) { }

	/// Returns a description to be shown in the history view.
  virtual QString GetStringDesc() const { return "abstract"; }

protected:
  const Person* owner;
  bool isToBeLogged;
  Timestamp virtualTime;
  Event* parentEvent;
  QList<Event*> childEvents;
  size_t level;
  bool operationRequired;

  virtual void Serialise( RakNet::BitStream& bs ) const;
  void Deserialise( QXmlStreamReader& xml, const iSession* session );
};

#endif
