/*
* log.cpp
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/

//====================================================================================
// QT Includes
//====================================================================================
#include <QtXml/QXmlStreamReader>
#include <QtXml/QXmlStreamWriter>
#include <QtGui/QTreeView>
#include <QtGui/QHeaderView>
//====================================================================================
// Project Includes
//====================================================================================
#include <typeinfo>
#include <vector>
#include <algorithm>
#include "network/messageids.h"
#include "network/network.h"
#include "raknet/BitStream.h"
#include "raknet/RakNetTypes.h"
#include "inetwork.h"
#include "isession.h"
//====================================================================================
// Local Includes
//====================================================================================
#include "chat.h"
#include "delete.h"
#include "draw.h"
#include "event.h"
#include "log.h"
#include "modify.h"
#include "timestamp.h"
//------------------------------------------------------------------------------------
// Forward Declarations
//------------------------------------------------------------------------------------
using namespace std;
Timestamp* Log::now	= NULL;
Log* Log::myself = NULL;

Log::Log() : testing(false), ownerInit(false), root(new Event()), currentEvent(root)
{
  owner = new Person(0, "You", SystemAddress(), 0, 0);
  Network::GetSingleton().RegisterPacketReciever(this);
  view = new QTreeView();
  view->setWindowFlags(Qt::Tool | Qt::WindowStaysOnTopHint);
  view->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOn);
  view->setModel(this);
  QHeaderView *h = view->header();
  h->setResizeMode(QHeaderView::ResizeToContents);
  QObject::connect(view,SIGNAL(clicked(const QModelIndex)),this,SLOT(OnEventClicked(const QModelIndex)));
  QObject::connect(view,SIGNAL(expanded(QModelIndex)),this,SLOT(expandedClicked(QModelIndex)));

  view->setExpandsOnDoubleClick(false);
}

Log::~Log()
{
  for( EventBag::iterator it = events.begin(); it != events.end(); ++it )
    delete *it;
}

void Log::Serialise( QXmlStreamWriter& xml ) const
{
  xml.writeStartElement( "log" );
  xml.writeAttribute( "size", QString::number( (uint)events.size() ) );

  for( EventBag::const_iterator it = events.begin(); it != events.end(); ++it )
    (*it)->Serialise( xml );

  for( EventBag::const_iterator it = chatEvents.begin(); it != chatEvents.end(); ++it )
    (*it)->Serialise( xml );

  xml.writeEndElement();
}

void Log::Deserialise( QXmlStreamReader& xml )
{
  Q_ASSERT( xml.isStartElement() && xml.name() == "log" );
  size_t size = xml.attributes().value("size").toString().toUInt();

  ClearLog();
  //events.reserve( size );
  while( !xml.atEnd() )
  {
    xml.readNext();

    if( xml.isEndElement() )
      break;

    if( xml.isStartElement() )
    {
      Event* e = NULL;
      if( xml.name() == "chat" )
        e = new Chat( xml, session );
      else if( xml.name() == "draw" )
        e = new Draw( xml, session );
      else if( xml.name() == "modify" )
        e = new Modify( xml, session );
      else if( xml.name() == "delete" )
        e = new Delete( xml, session );
      else
        ReadUnknownElement( xml );

      if( e )
        Add( e );
    }
  }
}

void Log::ReadUnknownElement( QXmlStreamReader& xml ) const
{
  Q_ASSERT( xml.isStartElement() );

  while( !xml.atEnd() )
  {
    xml.readNext();

    if( xml.isEndElement() )
      break;

    if( xml.isStartElement() )
      ReadUnknownElement( xml );
  }
}

void Log::ClearLog()
{
  // Delete all elements from the EventBags and reset the history view.
  for( EventBag::iterator it = events.begin(); it != events.end(); ++it )
    delete *it;
  events.clear();

  for( EventBag::iterator it = chatEvents.begin(); it != chatEvents.end(); ++it )
    delete *it;
  chatEvents.clear();

  currentEvent = root;
  root->ClearChildren();
  delete now;
  now = NULL;

  view->reset();
}

void Log::SetOwner(Person* logOwner)
{
  if(!ownerInit)
  {
    delete owner;
    owner = logOwner;
    ownerInit = true;
  }
}

void Log::Add(Event* event, bool logTemp)
{
  // If the event is a chat event then we keep it in a different bag.
  Chat* chat = dynamic_cast<Chat*>(event);
  if(chat)
  {
    chatEvents.push_back(chat);
  }
  else if(event->IsToBeLogged())
  {
    // Else insert into the log.
    Now().Update( event->GetTimestamp() );
    if( events.size() == 0 )
      events.push_back( event );
    else if( !( *events.back() > *event ) )
      events.push_back( event );
    else
    {
      EventBag::iterator it = events.end();
      EventBag::iterator next = it;
      --next;
      while( it != events.begin() && **next > *event )
      {
        --it;
        if( next != events.begin() )
          --next;
      }
      events.insert( it, event );
    }
  }
  else if(logTemp)
  {
    // If it's a temp event, push to the temp event bad.
    tempEvents.push_back( event );
  }

  // Set the root if not set.
  if( !chat && event->IsToBeLogged() && !root )
    root = event;

  // Execute the event.
  event->Execute();
  if(!chat && event->IsToBeLogged() && currentEvent == event->GetParentEvent())
  {
    // Set the current event.
    currentEvent = event;
  }

  if(!event->IsToBeLogged() && !logTemp)
  {
    delete event;
  }
  this->view->reset();
  this->view->expandAll();
}

void Log::JumpTo(Event* e, bool save)
{
  // Case 1: event is a (grand)parent of currentEvent.
  if(currentEvent && WalkUp(currentEvent, &Event::operator ==, e))
  {
    // Walk up from currentEvent to event, calling undo.
    WalkUp(currentEvent, &Event::Undo, e);
  }
  // Case 2: currentEvent is a (grand)parent of event.
  else if(WalkUp(e, &Event::operator ==, currentEvent))
  {
    // Walk up from event to currentEvent, marking that a redo is needed.
    WalkUp(e, &Event::MarkOperationRequired, currentEvent);

    // Walk down from currentEvent to event, calling redo.
    WalkDown(currentEvent, &Event::Redo);
  }
  // Case 3: event is in a different branch to currentEvent.
  else
  {
    // Find the common grandparent.
    Event* grandparent = FindGrandparent(e, currentEvent);

    // Walk up from currentEvent to grandparent, calling undo.
    WalkUp(currentEvent, &Event::Undo, grandparent);

    // Walk up from event to grandparent, marking that a redo is needed.
    WalkUp(e, &Event::MarkOperationRequired, grandparent);

    // Walk down from grandparent to event, calling redo.
    WalkDown(grandparent, &Event::Redo);
  }

  if(save)
  {
    jumpHistory.push_back(currentEvent);
  }

  currentEvent = e;
}

void Log::JumpBack()
{
  if(jumpHistory.size())
  {
    JumpTo(jumpHistory.at(jumpHistory.size()-1), false);
    jumpHistory.pop_back();
  }
}

Event* Log::FindGrandparent(Event* e1, Event* e2)
{
  // Get on the same tree level.
  while(e1->GetLevel() < e2->GetLevel())
  {
    e2 = e2->GetParentEvent();
  }

  while(e2->GetLevel() < e1->GetLevel())
  {
    e1 = e1->GetParentEvent();
  }

  // Walk up until they're the same event, then return this event.
  while(e1 != e2)
  {
    e1 = e1->GetParentEvent();
    e2 = e2->GetParentEvent();
  }

  return e1;
}

void Log::WalkUp(Event* e, void (Event::*operation)(void), Event* target)
{
  if(e != target)
  {
    (e->*operation)();
    WalkUp(e->GetParentEvent(), operation, target);
  }
}

bool Log::WalkUp(Event* event1, bool (Event::*operation)(const Event&) const, Event* event2) const
{
  bool result = !event2 || (event1->*operation)(*event2);
  if(!result && event1->GetParentEvent())
  {
    result |= WalkUp(event1->GetParentEvent(), operation, event2);
  }

  return result;
}

void Log::WalkDown(Event* e1, void (Event::*operation)(void))
{
  QList<Event*>::const_iterator itr = e1->GetChildren();
  for(int i = 0; i < e1->GetChildrenCount(); i++)
  {
    if(itr[i]->IsOperationRequired())
    {
      (itr[i]->*operation)();
      WalkDown(itr[i], operation);
      itr[i]->MarkOperationNotRequired();
    }
  }
}

void Log::RemoveTemp(Event* event)
{
  size_t index = 0;
  Event* temp = FindEvent(event->GetOwner(), index);
  if(temp)
  {
    TempDelete* d = new TempDelete((Draw*)temp);
    d->Execute();
    EventBag::iterator itr(tempEvents.begin() + index);
    delete *itr;
    tempEvents.erase(itr);
    delete d;
  }
}

bool Log::ReceivePacket(Packet* packet, unsigned char type)
{
  switch(type)
  {
  case ID_CHAT_EVENT:
    {
      RakNet::BitStream bs(packet->data, packet->length, false);

      MessageID type;
      bs.Read(type);
      unsigned char r;
      bs.Read(r);
      char order;
      bs.Read(order);

      Add(new Chat(bs, session));
      return true;
    }
  case ID_DRAW_EVENT:
    {
      RakNet::BitStream bs(packet->data, packet->length, false);

      MessageID type;
      bs.Read(type);
      unsigned char r;
      bs.Read(r);
      char order;
      bs.Read(order);

      Draw* draw = new Draw(bs, session);
      if(draw->IsToBeLogged())
      {
        RemoveTemp(draw);
      }

      Add(draw, true);
      return true;
    }
  case ID_DELETE_EVENT:
    {
      RakNet::BitStream bs(packet->data, packet->length, false);

      MessageID type;
      bs.Read(type);
      unsigned char r;
      bs.Read(r);
      char order;
      bs.Read(order);

      Add(new Delete(bs, session));
      return true;
    }
  case ID_MODIFY_EVENT:
    {
      RakNet::BitStream bs(packet->data, packet->length, false);

      MessageID type;
      bs.Read(type);
      unsigned char r;
      bs.Read(r);
      char order;
      bs.Read(order);

      Add(new Modify(bs, session));
      return true;
    }
  }

  return false;
}

Log::ChatIterator Log::begin()
{
  return ChatIterator( events.begin(), events.begin(), events.end() );
}

Log::ChatIterator Log::end()
{
  return ChatIterator( events.end(), events.begin(), events.end() );
}

Event* Log::FindEvent( const Timestamp& ts )
{
  if(events.size() == 0)
  {
    return root;
  }

  Event e(ts);
  EventBag::iterator pos = upper_bound( events.begin(), events.end(), &e, eventComparer );
  --pos;
  while( ! ( (*pos)->GetTimestamp() < ts ) )
  {
    if(ts == (*pos)->GetTimestamp())
    {
      return *pos;
    }

    if(pos == events.begin())
      break;

    --pos;
  }

  for( pos = events.begin(); pos != events.end(); ++pos )
  {
    if(ts == (*pos)->GetTimestamp())
      return *pos;
  }

  return root;
}

Event* Log::FindEvent( const Person* person, size_t& i )
{
  for(i=0; i<tempEvents.size(); i++)
  {
    if(person == tempEvents[i]->GetOwner())
    {
      return tempEvents[i];
    }
  }

  return NULL;
}

const Timestamp* Log::GetLastTimestamp()
{
  return &currentEvent->GetTimestamp();
}

Event* Log::DeserializeEventPointer( RakNet::BitStream& bs )
{
  Timestamp ts( bs, true );
  return FindEvent( ts );
}

void Log::ForwardEvents(iNetwork* network, SystemAddress& target, const Timestamp* ts)
{
  for(size_t i=0; i<events.size(); i++)
  {
    if(!ts || !(*ts > events[i]->GetTimestamp()))
      events[i]->SendEvent(network, HIGH_PRIORITY, &target);
  }
}

bool Log::EventCompareClass::operator()( const Event* e1, const Event* e2 ) const
{
  return *e1 < *e2;
}

void Log::OnHistoryClicked()
{
  view->show();
}

QModelIndex Log::index(int row, int column, const QModelIndex &parent) const
{
  if (!hasIndex(row, column, parent))
    return QModelIndex();

  Event *parentItem;

  if (!parent.isValid())
    parentItem = this->root;
  else
    parentItem = static_cast<Event*>(parent.internalPointer());

  Event *childItem = parentItem->GetChildren()[row];
  if (childItem)
    return createIndex(row, column, childItem);
  else
    return QModelIndex();
}
QModelIndex Log::parent(const QModelIndex &index) const
{
  if (!index.isValid())
    return QModelIndex();

  Event *childItem = static_cast<Event*>(index.internalPointer());
  Event *parentItem = childItem->GetParentEvent();

  if (parentItem == root)
    return QModelIndex();

  return createIndex(parentItem->row(), 0, parentItem);

}
int Log::rowCount(const QModelIndex &parent) const
{
  Event *parentItem;
  if (parent.column() > 0)
    return 0;

  if (!parent.isValid())
    parentItem = root;
  else
    parentItem = static_cast<Event*>(parent.internalPointer());

  if (root == NULL) return 0;
  return parentItem->GetChildrenCount();

}
int Log::columnCount(const QModelIndex &parent) const
{
  if (parent.isValid())
    return static_cast<Event*>(parent.internalPointer())->columnCount();
  else
    return root->columnCount();

}
QVariant Log::data(const QModelIndex &index, int role) const
{
  if (!index.isValid())
    return QVariant();

  if (role != Qt::DisplayRole)
    return QVariant();

  Event *item = static_cast<Event*>(index.internalPointer());

  return QString(item->GetStringDesc());
}

void Log::OnEventClicked(const QModelIndex &index){
  Event *item = static_cast<Event*>(index.internalPointer());
  JumpTo(item);
}
void Log::expandedClicked(const QModelIndex & index){
  view->resizeColumnToContents(index.column());
}
