/*
* log.h
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/
#ifndef __LOG_H__
#define __LOG_H__

//====================================================================================
// QT Includes
//====================================================================================

//====================================================================================
// Project Includes
//====================================================================================
#include <vector>
#include <typeinfo>
#include "ilog.h"
#include "ipacketreceiver.h"
#include "person.h"
#include "logger/logger.h"

//====================================================================================
// Local Includes
//====================================================================================
#include "timestamp.h"
#include <QtCore/QAbstractItemModel>

//------------------------------------------------------------------------------------
// Forward Declarations
//------------------------------------------------------------------------------------
class Event;
class Chat;
class TempModify;
class QTreeView;
struct iSession;
namespace RakNet
{
  class BitStream;
}
class QXmlStreamWriter;
class QXmlStreamReader;

/**
* Log keeps track of all events happening on the whiteboard. The can be ordered by
* by logical time (see Timestamp).
*/
class Log : public QAbstractItemModel, public iLog, public iPacketReceiver
{
  Q_OBJECT
    public slots:
      void OnHistoryClicked();
      void OnEventClicked(const QModelIndex &index);
      void expandedClicked(const QModelIndex & index);

public:
  Log();
  virtual ~Log();
  QModelIndex index(int row, int column, const QModelIndex &parent) const;
  QModelIndex parent(const QModelIndex &index) const;
  int rowCount(const QModelIndex &parent) const;
  int columnCount(const QModelIndex &parent) const;
  QVariant data(const QModelIndex &index, int role) const;



  static Log& GetSingleton()
  {
    if(!myself)
    {
      myself = new Log();
    }

    return *myself;
  }

  static Log* GetSingletonPtr()
  {
    return &GetSingleton();
  }

  /**
  * Provides the most recent timestamp we know about.
  */
  static Timestamp& Now()
  {
    if( !now )
    {
      Timestamp::Vectortime t;
      now = new Timestamp( t );
    }
    return *now;
  }

  /**
  * Set the owner of this log. This should happen as early as possible, because
  * the ticking of the local logical time depends of the ID of the owner.
  * @see Timestamp
  */
  void SetOwner( Person* logOwner );

  /**
  * Get the ID of the client this Log is running on.
  */
  inline Person* GetOwner()
  { return owner; }

  inline unsigned int GetClientID()
  {
    if(testing)
      return testID;
    return owner->GetID();
  }

  /**
  * Use for testing only.
  */
  inline void SetClientID(unsigned int id)
  {
    testing = true;
    testID = id;
  }

  /**
  * Add an Event to the log and synchonize the clock in this Log with it's Timestamp.
  * @see Now()
  * @see Timestamp.Update()
  */
  void Add(Event* event, bool logTemp = false);

  void JumpTo(Event* event, bool save = true);
  void JumpBack();

  /**
  * Clear the log.
  */
  void ClearLog();

  /**
  * Remove an event from tempEvents.
  */
  void RemoveTemp(Event* event);

  /**
  * Callback for handling packets. Implements iPacketReceiver.
  */
  virtual bool ReceivePacket(Packet* packet, unsigned char type);

  template <class T> class Iterator;
  typedef Iterator<Chat> ChatIterator;
  ChatIterator begin();
  ChatIterator end();

  inline void SetSession(iSession* session)
  {
    this->session = session;
  }

  Event* FindEvent( const Timestamp& ts );
  Event* FindEvent( const Person* person, size_t& index );
  Event* DeserializeEventPointer( RakNet::BitStream& bs );

  virtual void ForwardEvents(iNetwork* network, SystemAddress& target, const Timestamp* ts = 0);

  virtual void Serialise( QXmlStreamWriter& xml ) const;
  virtual void Deserialise( QXmlStreamReader& xml );

  inline Event* GetCurrentEvent() { return currentEvent; }

  const Timestamp* GetLastTimestamp();

private:
  /// Singleton object pointer.
  static Log* myself;
  QTreeView * view;

  static Timestamp* now;
  Person* owner;

  bool ownerInit;

  /// Unit test id.
  unsigned int testID;
  bool testing;

  iSession* session;

  typedef std::vector<Event*> EventBag;
  EventBag events;
  EventBag chatEvents;
  EventBag tempEvents;

  Event* root;
  Event* currentEvent;
  EventBag jumpHistory;

  class EventCompareClass
  {
  public:
    bool operator()( const Event* e1, const Event* e2 ) const;
  } eventComparer;

  // Find a common grandparent event of two events.
  Event* FindGrandparent(Event* e1, Event* e2);

  // For execution of recursive operations.
  void WalkUp(Event* e, void (Event::*operation)(void), Event* target);
  bool WalkUp(Event* event1, bool (Event::*operation)(const Event&) const, Event* event2) const;
  void WalkDown(Event* e, void (Event::*operation)(void));

  void ReadUnknownElement( QXmlStreamReader& xml ) const;
};

template <class T>
class Log::Iterator
{
public:
  Iterator( EventBag::iterator pos, EventBag::iterator head, EventBag::iterator tail );
  Iterator& operator++();
  Iterator& operator--();
  bool operator==( const Iterator<T>& it ) const;
  bool operator!=( const Iterator<T>& it ) const;
  const T& operator*() const;
  Iterator& operator=( const Iterator<T>& it );
private:
  EventBag::iterator pt;
  EventBag::iterator begin;
  EventBag::iterator end;
  const std::type_info& type;
};


// oh no it's a template!

template <class T>
Log::Iterator<T>::Iterator( EventBag::iterator pos, EventBag::iterator head, EventBag::iterator tail )
: pt( pos ), begin( head ), end( tail ), type( typeid( T ) )
{
  while( pt != end && typeid( **pt ) != type )
    ++pt;
}

template <class T>
Log::Iterator<T>& Log::Iterator<T>::operator++()
{
  if( pt == end )
    return *this;

  ++pt;
  while( pt != end && typeid( **pt ) != type )
    ++pt;
  return *this;
}

template <class T>
Log::Iterator<T>& Log::Iterator<T>::operator--()
{
  if( pt == begin )
    return *this;

  --pt;
  while( pt != begin && typeid( **pt ) != type )
    --pt;
  return *this;
}

template <class T>
bool Log::Iterator<T>::operator!=( const Iterator<T>& it ) const
{
  return this->pt != it.pt;
}

template <class T>
bool Log::Iterator<T>::operator==( const Iterator<T>& it ) const
{
  return this->pt == it.pt;
}

template <class T>
const T& Log::Iterator<T>::operator*() const
{
  return (const T&) **pt;
}

template <class T>
Log::Iterator<T>& Log::Iterator<T>::operator=( const Iterator<T>& it )
{
  this->pt = it.pt;
  this->begin = it.begin;
  this->end = it.end;

  return *this;
}


#endif


