/*
* modify.cpp
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/

//====================================================================================
// QT Includes
//====================================================================================
#include <QtXml/QXmlStreamWriter>
#include <QtXml/QXmlStreamReader>
//====================================================================================
// Project Includes
//====================================================================================
#include "network/messageids.h"
#include "raknet/BitStream.h"
#include "graphics/DescEllipse.h"
#include "graphics/DescLine.h"
#include "graphics/DescRect.h"
#include "graphics/DescText.h"
#include "graphics/DescPath.h"
#include "graphics/DescDeserialiser.h"
//====================================================================================
// Local Includes
//====================================================================================
#include "ilogcallback.h"
#include "draw.h"
#include "modify.h"
//------------------------------------------------------------------------------------
// Forward Declarations
//------------------------------------------------------------------------------------

iLogCallback* Modify::callback = NULL;


Modify::Modify(iDesc* desc, const Draw* drawEvent, bool isToBeLogged, bool hasDrawEvent)
: Event(isToBeLogged), desc(desc), draw(drawEvent), oldState(draw->GetDesc()), moving(hasDrawEvent)
{}

TempModify::TempModify(iDesc* desc, const Draw* drawEvent, bool resize)
: Modify(desc, drawEvent, false, resize)
{}

Modify::Modify(QPointF& point, const Draw* drawEvent, bool isToBeLogged)
: Event(isToBeLogged), desc(0), point(point), draw(drawEvent), moving(true)
{
}

TempModify::TempModify(QPointF& point, const Draw* drawEvent)
: Modify(point, drawEvent, false)
{
}

Modify::Modify(RakNet::BitStream& bs, const iSession* session)
: Event(bs, session)
{
  bs.ReadCompressed(moving);
  if(isToBeLogged || moving)
  {
    draw = (Draw*) Log::GetSingleton().DeserializeEventPointer( bs );
  }
  else
  {
    size_t unused;
    draw = (Draw*) Log::GetSingleton().FindEvent(GetOwner(), unused);
  }

  oldState = draw->GetDesc();

  bool edited;
  bs.ReadCompressed(edited);

  if(edited)
  {
    iDesc::DescType type;
    bs.ReadCompressed(type);

    switch(type)
    {
    case iDesc::Ellipse:
      {
        desc = new DescEllipse(bs);
        break;
      }
    case iDesc::Line:
      {
        desc = new DescLine(bs);
        break;
      }
    case iDesc::Rect:
      {
        desc = new DescRect(bs);
        break;
      }
    case iDesc::Text:
      {
        desc = new DescText(bs);
        break;
      }
    case iDesc::Freehand:
      {
        desc = new DescPath(bs);
        break;
      }
    }
  }
  else
  {
    desc = NULL;
    short x;
    short y;
    bs.ReadCompressed(x);
    bs.ReadCompressed(y);
    point.setX(x);
    point.setY(y);
  }
}

Modify::~Modify()
{
  //delete desc;
}

const Draw* Modify::GetDraw() const
{
  return draw;
}

void Modify::Execute() const
{
  if( callback )
    callback->ModifyEvent(*this);
}

void Modify::SendEvent(iNetwork* network, PacketPriority priority, SystemAddress* address)
{
  RakNet::BitStream bs;
  Serialise(bs);
  if(isToBeLogged)
  {
    network->SendBitStream(bs, priority, RELIABLE_ORDERED, 2, address);
  }
  else
  {
    network->SendBitStream(bs, priority, UNRELIABLE_SEQUENCED, 3, address);
  }
}

void Modify::Serialise( RakNet::BitStream& bs ) const
{
  bs.Write((MessageID)ID_MODIFY_EVENT);

  if(!isToBeLogged)
  {
    bs.Write((unsigned char)UNRELIABLE_SEQUENCED);
    bs.Write((char)3);
  }
  else
  {
    bs.Write((unsigned char)RELIABLE_ORDERED);
    bs.Write((char)2);
  }

  Event::Serialise( bs );

  bs.WriteCompressed(moving);
  if(isToBeLogged || moving)
  {
    draw->GetTimestamp().Serialise( bs );
  }

  if(desc)
  {
    bs.WriteCompressed(true);
    desc->Serialise(bs);
  }
  else
  {
    bs.WriteCompressed(false);
    short x = point.x();
    short y = point.y();
    bs.WriteCompressed(x);
    bs.WriteCompressed(y);
  }
}

void Modify::Serialise( QXmlStreamWriter& xml ) const
{
  xml.writeStartElement( "modify" );

  draw->GetTimestamp().Serialise( xml );
  Event::Serialise( xml );


  //  if(desc)
  //  {
  desc->Serialise(xml);
  //  }
  //  else
  //  {
  //    xml.writeStartElement( "point" );
  //		xml.writeAttribute( "x", QString::number( point.x() ) );
  //		xml.writeAttribute( "y", QString::number( point.y() ) );
  //		xml.writeEndElement();
  //  }

  xml.writeEndElement();
}

Modify::Modify( QXmlStreamReader& xml, const iSession* session )
{
  Q_ASSERT( xml.isStartElement() && xml.name() == "modify" );

  while( !xml.atEnd() )
  {
    xml.readNext();

    if( xml.isEndElement() )
      break;

    if( xml.isStartElement() )
    {
      if( xml.name() == "event" )
      {
        Event::Deserialise( xml, session );
      }
      //			else if( xml.name() == "point" )
      //			{
      //				qreal x = xml.attributes().value( "x" ).toString().toDouble();
      //				qreal y = xml.attributes().value( "y" ).toString().toDouble();
      //				point = QPointF( x, y );
      //
      //				while( !xml.atEnd() && !xml.isEndElement() )
      //					xml.readNext();
      //			}
      else if( xml.name() == "desc" )
        desc = DescDeserialiser::Deserialise( xml );
      else if( xml.name() == "timestamp" )
      {
        Timestamp ts( xml );
        draw = (Draw*) Log::GetSingleton().FindEvent( ts );
      }
    }
  }
}

void Modify::Undo()
{
  iDesc* temp = desc;
  desc = oldState;
  oldState = temp;
  Execute();
}

QString Modify::GetStringDesc() const
{
  return QString( "modify " ) + draw->GetDesc()->GetStringDesc();
}

