/*
* timestamp.cpp
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/

//====================================================================================
// QT Includes
//====================================================================================
#include <QtXml/QXmlStreamWriter>
#include <QtXml/QXmlStreamReader>

//====================================================================================
// Project Includes
//====================================================================================
#include <vector>
#include <iostream>
#include "raknet/BitStream.h"
#include "logger/logger.h"

//====================================================================================
// Local Includes
//====================================================================================
#include "log.h"
#include "timestamp.h"
//------------------------------------------------------------------------------------
// Forward Declarations
//------------------------------------------------------------------------------------


Timestamp::Timestamp()
{
  // Init as invalid.
  time.push_back(-1);
}

Timestamp::Timestamp( const Vectortime& t ) : time( t )
{}

Timestamp::Timestamp( RakNet::BitStream& bs, bool hasTimeStamp )
{
  if(!hasTimeStamp)
  {
    bs.ReadCompressed(hasTimeStamp);
  }

  if(hasTimeStamp)
  {
		// copy data directly into the vector internal space (fast)
    Vectortime::size_type length;
    bs.ReadCompressed(length);
    time.resize( length );
    bs.Read( (char*) &time[0], (const unsigned int) ( length*sizeof(Localtime) ) );
  }
  else
  {
    time.push_back(-1);
  }
}

Timestamp::~Timestamp()
{}

void Timestamp::Update( const Timestamp& ts )
{
	// use a componentwise maximum operator
  for( unsigned int i = 0; i < ts.time.size(); ++i )
  {
		// ts.time might be longer than this->time, so we have to append
    if( i >= time.size() )
      time.push_back( ts.time[i] );
    else if( ts.time[i] > time[i] )
      time[i] = ts.time[i];
  }
}

Timestamp& Timestamp::operator++()
{
	// make sure the current timestamp is big enough
  if( time.size() <= Log::GetSingleton().GetClientID() )
    time.resize( Log::GetSingleton().GetClientID()+1, 0 );

	// then increase the component this client is associated with
  ++time[ Log::GetSingleton().GetClientID() ];
  return *this;
}

bool Timestamp::operator<( const Timestamp& ts ) const
{
  return Compare( ts ) == earlier;
}

bool Timestamp::operator>( const Timestamp& ts ) const
{
  return Compare( ts ) == later;
}

bool Timestamp::operator==( const Timestamp& ts ) const
{
  return Compare( ts ) == same;
}

bool Timestamp::operator!=( const Timestamp& ts ) const
{
  return Compare( ts ) != same;
}

bool Timestamp::operator||( const Timestamp& ts ) const
{
  return Compare( ts ) == concurrent;
}

Timestamp::CompareType Timestamp::Compare( const Timestamp& ts ) const
{
  bool smaller = false;
  bool greater = false;
  bool equal = false;

  Vectortime::const_iterator it1 = time.begin();
  Vectortime::const_iterator it2 = ts.time.begin();
	// compare both timestamps componentwise
  while( it1 != time.end() || it2 != ts.time.end() )
  {
		// assume "missing" components are zero (new clients might have joined)
    Localtime t1 = 0;
    if( it1 != time.end() )
    {
      t1 = *it1;
      ++it1;
    }

    Localtime t2 = 0;
    if( it2 != ts.time.end() )
    {
      t2 = *it2;
      ++it2;
    }

		// save the difference
    if( t1 < t2 )
      smaller = true;
    else if( t1 == t2 )
      equal = true;
    else
      greater = true;

    if( smaller && equal && greater )
      return concurrent;
  }

	// all elements are the same
  if     ( !smaller && !greater && equal )
    return same;
	// all elements are less or equal
  else if( smaller && !greater )
    return earlier;
	// all elements are greater or equal
  else if( !smaller &&  greater )
    return later;
  else
    return concurrent;
}

std::ostream& operator<<( std::ostream& os, const Timestamp& ts )
{
	// simply join the components by commas
  os << "[";
  for( Timestamp::Vectortime::const_iterator it = ts.time.begin();
    it != ts.time.end(); ++it )
	{
    os << ( (it == ts.time.begin()) ? "" : "," ) << *it;
	}
  os << "]";
  return os;
}

void Timestamp::Serialise( RakNet::BitStream& bs ) const
{
	// copy data directly from the internal vector memory into the bitstream (fast)
  bs.WriteCompressed( (unsigned int)time.size() );
  bs.Write( (const char*) &time[0], (const unsigned int) ( time.size()*sizeof(Localtime) ) );
}

void Timestamp::Serialise( QXmlStreamWriter& xml ) const
{
  xml.writeStartElement("timestamp");
  xml.writeAttribute( "size", QString::number( (uint)time.size() ) );

	// write each component
  for( Vectortime::const_iterator it = time.begin(); it != time.end(); ++it )
  {
    xml.writeStartElement("localtime");
    xml.writeAttribute( "value", QString::number( *it ) );
    xml.writeEndElement();
  }

  xml.writeEndElement();
}

Timestamp::Timestamp( QXmlStreamReader& xml )
{
  Q_ASSERT( xml.isStartElement() && xml.name() == "timestamp" );

	// allocate enough memory in time vector
  size_t size = xml.attributes().value( "size" ).toString().toUInt();
  time.clear();
  time.reserve( size );

  size_t i = 0;
  while( !xml.atEnd() )
  {
    xml.readNext();

    if( xml.isEndElement() )
      break;

    if( xml.isStartElement() )
    {
			// read each component
      Q_ASSERT( xml.name() == "localtime" );
      time.push_back( xml.attributes().value("value").toString().toUInt() );
      while( !xml.atEnd() && !xml.isEndElement() )
        xml.readNext();
    }
  }
}


