/*
* timestamp.h
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/
#ifndef __TIMESTAMP_H__
#define __TIMESTAMP_H__

//====================================================================================
// QT Includes
//====================================================================================

//====================================================================================
// Project Includes
//====================================================================================
#include <vector>
#include <iostream>

//====================================================================================
// Local Includes
//====================================================================================

//------------------------------------------------------------------------------------
// Forward Declarations
//------------------------------------------------------------------------------------
class QXmlStreamWriter;
class QXmlStreamReader;
namespace RakNet
{
  class BitStream;
}

/**
* Timestamp implements a timestamp of a vector clock, keeping track of changes
* in the network (according to Leslie Lamport and Colin J. Fidge).
* This Timestamps can safely be used as Event-IDs as they are unique.
*/
class Timestamp
{
public:
  typedef unsigned int Localtime;
  typedef std::vector<Localtime> Vectortime;

public:
  /// Creates an invalid Timestamp.
	Timestamp();

	/// Creates a timestamp as represented in the passed STL-vector.
  Timestamp( const Vectortime& t );


	/**
	 * Deserialises a timestamp from the network.
	 * @argument bs The network bitstream to read from.
	 * @argument hasTimeStamp If set to true an in valid timestamp will be created.
	 *                        This argument is simply for convenience of recursive
	 *                        deserialisation.
	 */
  Timestamp( RakNet::BitStream& bs, bool hasTimeStamp = true );

	/**
	 * Deserialises a timestamp from the network.
	 * @argument xml A qt xml stream. The current item in the stream should be a
	 *               timestamp tag.
	 */
  Timestamp( QXmlStreamReader& xml );

  ~Timestamp();

  /**
  * Will set the timestamp to a time equal or later to the time currently stored
  * and the Timestamp ts.
  * @argument ts The Timestamp to synchronize with.
  */
  void Update( const Timestamp& ts );

  /**
  * Let the clock tick once.
  */
  Timestamp& operator++();

  /**
  * Checks whether this timestamp happened earlier than ts.
  */
  bool operator<( const Timestamp& ts ) const;

  /**
  * Checks whether this timestamp happened later than ts.
  */
  bool operator>( const Timestamp& ts ) const;

  /**
  * Checks whether this timestamp is the same as ts.
  */
  bool operator==( const Timestamp& ts ) const;

  /**
  * Checks whether this timestamp is different same to ts.
  */
  bool operator!=( const Timestamp& ts ) const;

  /**
  * Checks whether this timestamp is concurrent to ts.
  */
  bool operator||( const Timestamp& ts ) const;

	/// For convenience of debugging output, this function writes the timestamp to an STL out stream.
  friend std::ostream& operator<<( std::ostream& os, const Timestamp& ts );


	/**
	 * Serialises this timestamp to the network.
	 * @argument bs The network bitstream to write to.
	 */
  void Serialise( RakNet::BitStream& bs ) const;

	/**
	 * Serialises this timestamp to qt XML stream.
	 * @argument xml The XML stream to write to.
	 */
  void Serialise( QXmlStreamWriter& xml ) const;

private:
  Vectortime time;

  enum CompareType
  {
    earlier,
    later,
    concurrent,
    same
  };

	/// All compare operators depend on this compare implementation.
  CompareType Compare( const Timestamp& ts ) const;
};

#endif

