/*
* logger.h
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/
#ifndef __LOGGER_H__
#define __LOGGER_H__

//====================================================================================
// QT Includes
//====================================================================================

//====================================================================================
// Project Includes
//====================================================================================
#include <iostream>
#include <string>
#include <list>

//====================================================================================
// Local Includes
//====================================================================================

//------------------------------------------------------------------------------------
// Forward Declarations
//------------------------------------------------------------------------------------
class QString;

/** Logger recieves error, logging and debug messages and pipes them to ofstreams.
*
* Logger is holding a list of output streams to where it will forward the log messages
* it receives. Each of those output streams has a verbosity level to control how
* much messages will be sent there, so this class can be used both for debugging
* and normal logging purposes.
*
* Messages sent to the logger have a verbosity level as well. The logger will prefix
* the message with a timestamp and write it to those output streams having an equal
* or greater verbosity level.
*
* Verbosity convensions:
* 20  - errors
* 40  - logging
* 60  - debugging
* 80  - extreme debugging
* 100 - insane debugging
*/
class Logger
{
public:
  enum VerbosityLevels
  {
    ErrorLevel = 20,
    LoggingLevel = 40,
    DebugLevel = 60,
    ExtremeDebugLevel = 80,
    InsaneDebugLevel = 100
  };

  static Logger& getSingleton()
  {
    if( ! singleton )
      singleton = new Logger;

    return *singleton;
  }

  Logger();
  ~Logger();

  /**
  * Add an output destination to the Logger.
  * @param outstream stream to send log messages to
  * @param verbosityLevel messages with a verbosity level exceeding this value will be suppressed
  */
  void add( std::ostream& outstream, int verbosityLevel );

  /**
  * Open a file for writing (no appending) and use it as output destination.
  * @param filename file to open
  * @param verbosityLevel messages with a verbosity level exceeding this value will be suppressed
  */
  void add( const std::string& filename, int verbosityLevel );

  /**
  * Send a log message to the output destination added so far. The Logger will prefix
  * the message with a timestamp and send it to the outputs which have an equal or
  * higher verbosity level.
  * @param verbosityLevel verbosity level of the message
  * @param message message which will be sent
  */
  void log( int verbosityLevel, std::string message );
  void log( int verbosityLevel, QString message );
  void log( int verbosityLevel, const char* message );
protected:
private:
  static Logger* singleton;

  class Output;
  std::list<Output> out;
  std::list<std::ostream*> toFree;
  static const int timeStampLength;

  std::string getTimeStamp();
};

/** Output is a (outstream, verbosity) tuple which is used by the Logger.
*/
class Logger::Output
{
public:
  Output( std::ostream& outstream, int verbosityLevel );
  ~Output();

  void log( int verbosityLevel, std::string message );

private:
  int verbosity;
  std::ostream& stream;
};

#define DEBUG

#ifdef DEBUG
#define DBG(lvl,msg) Logger::getSingleton().log( lvl, msg )
#else
#define DBG(lvl,msg) ;
#endif

#endif
