/*
* messages.cpp
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/

//====================================================================================
// QT Includes
//====================================================================================

//====================================================================================
// Project Includes
//====================================================================================
#include "logger/logger.h"
#include "log/timestamp.h"
#include "raknet/StringCompressor.h"

//====================================================================================
// Local Includes
//====================================================================================
#include "messages.h"
#include "messageids.h"
#include "network.h"

//------------------------------------------------------------------------------------
// Forward Declarations and Preprocessor Definitions
//------------------------------------------------------------------------------------
using namespace RakNet;
using namespace std;

/************************************************************************************/

MSG_PEER_AUTHENTICATE::MSG_PEER_AUTHENTICATE(Packet* packet) : bs(packet->data, packet->length, false)
{
  MessageID type;
  bs.Read(type);

  bs.ReadCompressed(connecting);

  if(connecting)
  {
    bs.ReadCompressed(result);

    if(result == SUCCESSFUL)
    {
      bs.ReadCompressed(id);
      bs.ReadCompressed(myID);

      int length;
      bs.ReadCompressed(length);

      char* uname = new char[length];
      stringCompressor->DecodeString(uname, length, &bs);

      username = uname;
      delete uname;

      bs.ReadCompressed(address.binaryAddress);
      bs.ReadCompressed(address.port);
      bs.ReadCompressed(localTime);

      size_t size;
      bs.ReadCompressed(size);

      for(size_t i=0; i<size; i++)
      {
        short id;
        bs.ReadCompressed(id);
        connectedIDs.push_back(id);
      }
    }
  }
  else
  {
    bs.ReadCompressed(clientVersion);

    int ulength;
    bs.ReadCompressed(ulength);

    char* uname = new char[ulength];
    stringCompressor->DecodeString(uname, ulength, &bs);

    int plength;
    bs.ReadCompressed(plength);

    char* pass = new char[plength];
    stringCompressor->DecodeString(pass, plength, &bs);

    username = uname;
    password = pass;

    delete uname;
    delete pass;

    bs.ReadCompressed(address.binaryAddress);
    bs.ReadCompressed(address.port);
    bs.ReadCompressed(localTime);
  }
}

MSG_PEER_AUTHENTICATE::MSG_PEER_AUTHENTICATE(const char* username, const char* password, SystemAddress& address, short localTime)
  : username(username), password(password), address(address), localTime(localTime)
{
}

MSG_PEER_AUTHENTICATE::MSG_PEER_AUTHENTICATE(LOGIN_RESULT result, const short myID, const short id, const char* username, 
                                             SystemAddress& address, short localTime, const std::vector<short>& connectedIDs)
{
  this->result = result;
  
  if(result == SUCCESSFUL)
  {
    this->id = myID;
    this->myID = id;
    this->username = username;
    this->address = address;
    this->localTime = localTime;
    this->connectedIDs = connectedIDs;
  }
}

BitStream& MSG_PEER_AUTHENTICATE::GetResponderBitStream()
{
  bs.Write((MessageID)ID_PEER_AUTHENTICATE);
  bs.WriteCompressed(true);
  bs.WriteCompressed(result);

  if(result == SUCCESSFUL)
  {
    bs.WriteCompressed(id);
    bs.WriteCompressed(myID);
    bs.WriteCompressed((unsigned int)username.GetLength()+1);
    stringCompressor->EncodeString(username, (int)username.GetLength()+1, &bs);

    bs.WriteCompressed(address.binaryAddress);
    bs.WriteCompressed(address.port);
    bs.WriteCompressed(localTime);
    bs.WriteCompressed((unsigned int)connectedIDs.size());
    for(size_t i=0; i<connectedIDs.size(); i++)
    {
      bs.WriteCompressed(connectedIDs[i]);
    }
  }

  return bs;
}

BitStream& MSG_PEER_AUTHENTICATE::GetConnectingBitStream()
{
  bs.Write((MessageID)ID_PEER_AUTHENTICATE);
  bs.WriteCompressed(false);
  bs.WriteCompressed((uint)NETVERSION);
  bs.WriteCompressed((unsigned int)username.GetLength()+1);
  stringCompressor->EncodeString(username, (int)username.GetLength()+1, &bs);
  bs.WriteCompressed((unsigned int)password.GetLength()+1);
  stringCompressor->EncodeString(password, (int)password.GetLength()+1, &bs);
  bs.WriteCompressed(address.binaryAddress);
  bs.WriteCompressed(address.port);
  bs.WriteCompressed(localTime);
  return bs;
}

/************************************************************************************/

MSG_PEER_RECONNECT::MSG_PEER_RECONNECT(Packet* packet) : bs(packet->data, packet->length, false)
{
  MessageID type;
  bs.Read(type);

  bs.ReadCompressed(reconnecting);
  if(reconnecting)
  {
    bs.ReadCompressed(result);

    if(result == SUCCESSFUL)
    {
      bs.ReadCompressed(id);

      size_t size;
      bs.ReadCompressed(size);

      for(size_t i=0; i<size; i++)
      {
        short id;
        bs.ReadCompressed(id);
        connectedIDs.push_back(id);
      }
    }
  }
  else
  {
    bs.ReadCompressed(id);

    int ulength;
    bs.ReadCompressed(ulength);

    char* uname = new char[ulength];
    stringCompressor->DecodeString(uname, ulength, &bs);

    int plength;
    bs.ReadCompressed(plength);

    char* pass = new char[plength];
    stringCompressor->DecodeString(pass, plength, &bs);

    username = uname;
    password = pass;

    delete uname;
    delete pass;

    ts = new Timestamp(bs, true);
  }
}

MSG_PEER_RECONNECT::MSG_PEER_RECONNECT(LOGIN_RESULT result, short id, const std::vector<short>& connectedIDs)
  : result(result), id(id), connectedIDs(connectedIDs), reconnecting(true)
{
}

MSG_PEER_RECONNECT::MSG_PEER_RECONNECT(short id, const char* username, const char* password, const Timestamp* ts)
: id(id), username(username), password(password), reconnecting(false), ts(ts)
{
}

RakNet::BitStream& MSG_PEER_RECONNECT::GetResponderBitStream()
{
  bs.Write((MessageID)ID_PEER_RECONNECT);
  bs.WriteCompressed(reconnecting);
  bs.WriteCompressed(result);
  if(result == SUCCESSFUL)
  {
    bs.WriteCompressed(id);
    bs.WriteCompressed((unsigned int)connectedIDs.size());
    for(size_t i=0; i<connectedIDs.size(); i++)
    {
      bs.WriteCompressed(connectedIDs[i]);
    }
  }

  return bs;
}

RakNet::BitStream& MSG_PEER_RECONNECT::GetReconnectingBitStream()
{
  bs.Write((MessageID)ID_PEER_RECONNECT);
  bs.WriteCompressed(reconnecting);
  bs.WriteCompressed(id);
  bs.WriteCompressed((unsigned int)username.GetLength()+1);
  stringCompressor->EncodeString(username, (int)username.GetLength()+1, &bs);
  bs.WriteCompressed((unsigned int)password.GetLength()+1);
  stringCompressor->EncodeString(password, (int)password.GetLength()+1, &bs);
  ts->Serialise(bs);
  return bs;
}

/************************************************************************************/

MSG_PROPOSE_PEER::MSG_PROPOSE_PEER(Packet* packet) : bs(packet->data, packet->length, false), encoded(false)
{
  MessageID type;
  bs.Read(type);

  bs.ReadCompressed(fromProposer);

  if(fromProposer)
  {
    bs.ReadCompressed(id);

    int ulength;
    bs.ReadCompressed(ulength);

    char* uname = new char[ulength];
    stringCompressor->DecodeString(uname, ulength, &bs);

    username = uname;

    delete uname;
  }
  else
  {
    bs.ReadCompressed(id);
    bs.ReadCompressed(accepted);
    if(accepted == ALSO_PROPOSING)
    {
      bs.ReadCompressed(pid);
    }
  }
}

RakNet::BitStream& MSG_PROPOSE_PEER::GetProposerBitStream()
{
  if(!encoded)
  {
    bs.Write((MessageID)ID_PROPOSE_PEER);
    bs.WriteCompressed(fromProposer);
    bs.WriteCompressed(id);
    bs.WriteCompressed((unsigned int)username.GetLength()+1);
    stringCompressor->EncodeString(username, (int)username.GetLength()+1, &bs);
  }

  return bs;
}

RakNet::BitStream& MSG_PROPOSE_PEER::GetAcceptorBitStream()
{
  if(!encoded)
  {
    bs.Write((MessageID)ID_PROPOSE_PEER);
    bs.WriteCompressed(fromProposer);
    bs.WriteCompressed(id);
    bs.WriteCompressed(accepted);
    if(accepted == ALSO_PROPOSING)
    {
      bs.WriteCompressed(pid);
    }
  }

  return bs;
}

/************************************************************************************/

MSG_NEW_PEER::MSG_NEW_PEER(Packet* packet) : bs(packet->data, packet->length, false), encoded(false)
{
  MessageID type;
  bs.Read(type);

  bs.ReadCompressed(id);

  bs.ReadCompressed(address.binaryAddress);
  bs.ReadCompressed(address.port);
  bs.ReadCompressed(peerTime);

  int ulength;
  bs.ReadCompressed(ulength);

  char* uname = new char[ulength];
  stringCompressor->DecodeString(uname, ulength, &bs);

  username = uname;
  delete uname;
}

MSG_NEW_PEER::MSG_NEW_PEER(const short id, const SystemAddress& address, const RakNet::RakString& username, const short peerTime)
: id(id), address(address), username(username), encoded(false), peerTime(peerTime)
{
}

RakNet::BitStream& MSG_NEW_PEER::GetBitStream()
{
  if(encoded)
  {
    return bs;
  }
  else
  {
    bs.Write((MessageID)ID_NEW_PEER);
    bs.WriteCompressed(id);
    bs.WriteCompressed(address.binaryAddress);
    bs.WriteCompressed(address.port);
    bs.WriteCompressed(peerTime);
    bs.WriteCompressed((unsigned int)username.GetLength()+1);
    stringCompressor->EncodeString(username, (int)username.GetLength()+1, &bs);
    encoded = true;
    return bs;
  }
}

/************************************************************************************/

MSG_LOST_PEER::MSG_LOST_PEER(Packet* packet) : bs(packet->data, packet->length, false), encoded(false)
{
  MessageID type;
  bs.Read(type);

  bs.ReadCompressed(address.binaryAddress);
  bs.ReadCompressed(address.port);
}

MSG_LOST_PEER::MSG_LOST_PEER(const SystemAddress& address) : address(address), encoded(false)
{
}

RakNet::BitStream& MSG_LOST_PEER::GetBitStream()
{
  if(encoded)
  {
    return bs;
  }
  else
  {
    bs.Write((MessageID)ID_LOST_PEER);
    bs.WriteCompressed(address.binaryAddress);
    bs.WriteCompressed(address.port);
    encoded = true;
    return bs;
  }
}
