/*
* messages.h
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/

#ifndef __MESSAGES_H__
#define __MESSAGES_H__

//====================================================================================
// QT Includes
//====================================================================================

//====================================================================================
// Project Includes
//====================================================================================
#include "log/chat.h"
#include "raknet/BitStream.h"
#include "raknet/RakString.h"

//====================================================================================
// Local Includes
//====================================================================================

//------------------------------------------------------------------------------------
// Forward Declarations and Preprocessor Definitions
//------------------------------------------------------------------------------------
struct Packet;
typedef unsigned int uint;

enum Accepted
{
  ACCEPTED,
  ALSO_PROPOSING,
  BAD_USERNAME,
};

enum LOGIN_RESULT
{
  NONE = 0,
  INCOMPATIBLE,
  UNAUTHORISED,
  ALREADYIN,
  SUCCESSFUL
};

class MSG_PEER_AUTHENTICATE
{
public:
  MSG_PEER_AUTHENTICATE(Packet* packet);
  MSG_PEER_AUTHENTICATE(const char* username, const char* password, SystemAddress& address, short localTime);
  MSG_PEER_AUTHENTICATE(LOGIN_RESULT result, const short myID, const short id, const char* username,
    SystemAddress& address, short localTime, const std::vector<short>& connectedIDs);

  const short& GetID() const { return id; }
  const short& GetMyID() const { return myID; }
  const char* GetUsername() const { return username; }
  const char* GetPassword() const { return password; }
  short GetTime() const { return localTime; }
  const bool& IsConnecting() const { return connecting; }
  const LOGIN_RESULT& GetResult() const { return result; }
  const uint& GetClientVersion() const { return clientVersion; }
  const SystemAddress& GetSystemAddress() const { return address; }
  const std::vector<short>& GetConnectedIDs() { return connectedIDs; }

  RakNet::BitStream& GetConnectingBitStream();
  RakNet::BitStream& GetResponderBitStream();

private:
  short id;
  short myID;
  uint clientVersion;
  bool connecting;
  RakNet::RakString username;
  RakNet::RakString password;
  SystemAddress address;
  short localTime;
  LOGIN_RESULT result;
  std::vector<short> connectedIDs;
  RakNet::BitStream bs;
};

class MSG_PEER_RECONNECT
{
public:
  MSG_PEER_RECONNECT(Packet* packet);
  MSG_PEER_RECONNECT(short id, const char* username, const char* password, const Timestamp* ts);
  MSG_PEER_RECONNECT(LOGIN_RESULT result, short id, const std::vector<short>& connectedIDs);

  const short& GetID() const { return id; }
  const char* GetUsername() const { return username; }
  const char* GetPassword() const { return password; }
  const bool& IsReconnecting() const { return reconnecting; }
  const LOGIN_RESULT& GetResult() const { return result; }
  const std::vector<short>& GetConnectedIDs() { return connectedIDs; }
  const Timestamp* GetTimestamp() { return ts; }

  RakNet::BitStream& GetReconnectingBitStream();
  RakNet::BitStream& GetResponderBitStream();

private:
    short id;
    RakNet::RakString username;
    RakNet::RakString password;
    const Timestamp* ts;
    LOGIN_RESULT result;
    std::vector<short> connectedIDs;
    bool reconnecting;
    RakNet::BitStream bs;
};

class MSG_PROPOSE_PEER
{
public:
  MSG_PROPOSE_PEER(Packet* packet);
  MSG_PROPOSE_PEER(short id, const RakNet::RakString& username) : id(id), username(username), fromProposer(true), encoded(false) { }
  MSG_PROPOSE_PEER(short id, short pid, Accepted accepted) : id(id), pid(pid), accepted(accepted), fromProposer(false), encoded(false) { }

  const short& GetID() const { return id; }
  const short& GetProposerID() const { return pid; }
  const RakNet::RakString& GetUsername() const { return username; }
  const Accepted& GetAccepted() const { return accepted; }
  const bool& IsFromProposer() const { return fromProposer; }

  RakNet::BitStream& GetProposerBitStream();
  RakNet::BitStream& GetAcceptorBitStream();

private:
  short id;
  short pid;
  RakNet::RakString username;
  Accepted accepted;
  RakNet::BitStream bs;
  bool fromProposer;
  bool encoded;
};

class MSG_NEW_PEER
{
public:
  MSG_NEW_PEER(Packet* packet);
  MSG_NEW_PEER(const short id, const SystemAddress& address, const RakNet::RakString& username, const short peerTime);

  const short& GetID() const { return id; }
  const RakNet::RakString& GetUsername() const { return username; }
  const SystemAddress& GetAddress() const { return address; }
  const short GetTime() const { return peerTime; }

  RakNet::BitStream& GetBitStream();

private:
  short id;
  SystemAddress address;
  RakNet::RakString username;
  short peerTime;
  RakNet::BitStream bs;
  bool encoded;
};

class MSG_LOST_PEER
{
public:
  MSG_LOST_PEER(Packet* packet);
  MSG_LOST_PEER(const SystemAddress& address);

  const SystemAddress& GetAddress() const { return address; }

  RakNet::BitStream& GetBitStream();

private:
  SystemAddress address;
  RakNet::BitStream bs;
  bool encoded;
};

#endif // __MESSAGES_H__
