/*
* network.cpp
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/

//====================================================================================
// QT Includes
//====================================================================================
#include <QtCore/QTime>
#include <QtCore/QTimer>

//====================================================================================
// Project Includes
//====================================================================================
#ifndef WIN32
#include <arpa/inet.h>
#endif

#include "ilog.h"
#include "isession.h"
#include "network/messageids.h"
#include "network/messages.h"
#include "raknet/RakNetworkFactory.h"
#include "raknet/RakPeerInterface.h"
#include "logger/logger.h"

//====================================================================================
// Local Includes
//====================================================================================
#include "network.h"

//------------------------------------------------------------------------------------
// Forward Declarations and Preprocessor Definitions
//------------------------------------------------------------------------------------
Network* Network::myself = NULL;
using namespace RakNet;

Network::Network() : connectActive(false), usedIds(0), reconnecting(false)
{
  // Init network.
  network = RakNetworkFactory::GetRakPeerInterface();

  // Register us on the QT event loop.
  QTimer* t = new QTimer(this);
  connect(t, SIGNAL(timeout()), SLOT(Process()));
  t->start(0);
}

Network::~Network()
{
  // Disconnect from all peers.
  network->Shutdown(5000);

  // Destroy.
  RakNetworkFactory::DestroyRakPeerInterface(network);
}

void Network::Connect(int lPort, int connectingPort, const char* ip, const char* username, const char* password, bool connecting)
{
  // Set some data.
  this->username = username;
  this->password = password;
  this->ip = ip;

  // If a connection already exists, disconnect.
  if(connectActive)
  {
    network->Shutdown(5000);
    connectActive = false;
  }

  // Set the port to listen on for new connections.
  listeningPort = lPort - 1;
  bool listening = false;

  // Start the network.
  while(!listening)
  {
    SocketDescriptor socket(++listeningPort, 0);
    listening = network->Startup(10, 30, &socket, 1);
  }

  // Set big limit - why is one needed? That's just how it is.
  network->SetMaximumIncomingConnections(1000);

  // If we're connecting to another peer (we might be just listening, as is done on session start)....
  if(connecting)
  {
    // Then attempt to connect to that peer.
    connectActive = network->Connect(ip, connectingPort, 0, 0);
    if(!connectActive)
    {
	  	DBG(0,QString("Failed to connect to peer at: %s:%d\n"));

      network->Shutdown(5000);
      return;
    }
  }
  else
  {
    // Else mark that our username isn't in use.
    usedUsernames.push_back(username);
    connectActive = true;
  }

  // Set a 5 second peer timeout.
  network->SetTimeoutTime(5000, UNASSIGNED_SYSTEM_ADDRESS);
}

void Network::Process()
{
  // If we're connected, process one packet.
  if(connectActive)
  {
    Packet* packet = network->Receive();
    if(packet)
    {
      ProcessPacket(packet, packet->data[0]);
      network->DeallocatePacket(packet);
    }
  }
}

void Network::ProcessPacket(Packet* &packet, unsigned char type)
{
	DBG(60, "Network::ProcessPacket processing packet...");

  switch(type)
  {
  case ID_CONNECTION_ATTEMPT_FAILED:
    {
      // If this was a reconnect attempt...
      if(reconnectOrder.size() != 0)
      {
        // Attempt to connect to another.
        char address[16];
        reconnectOrder.begin()[0]->GetAddress().ToString(false, address);
        bool connectSent = network->Connect(address, listeningPort, 0, 0);
        reconnectOrder.erase(reconnectOrder.begin());
      }
      else
      {
        // Else throw an error and shut down.
        DBG(0,QString("Failed to connect to peer."));
        reconnecting = false;
        connectActive = false;
        network->Shutdown(5000);
      }
      break;
    }
  case ID_CONNECTION_REQUEST_ACCEPTED:
    {
      if(!reconnecting)
      {
        // Send authentication request.
        MSG_PEER_AUTHENTICATE msg(username, password, packet->systemAddress, GetLocalTime());
        SendBitStream(msg.GetConnectingBitStream(), HIGH_PRIORITY, RELIABLE_ORDERED, 0, &packet->systemAddress);
      }
      else
      {
        // Send reconnect request.
        reconnecting = false;
        MSG_PEER_RECONNECT msg(log->GetOwner()->GetID(), username, password, log->GetLastTimestamp());
        SendBitStream(msg.GetReconnectingBitStream(), HIGH_PRIORITY, RELIABLE_ORDERED, 0, &packet->systemAddress);
      }

      break;
    }
  case ID_NO_FREE_INCOMING_CONNECTIONS:
    {
      /// Shutdown in this case.. happens if there's >1000 peers :)
      connectActive = false;
      network->Shutdown(5000);
      break;
    }
  case ID_DISCONNECTION_NOTIFICATION:
  case ID_CONNECTION_LOST:
    {
      // A peer has disconnected. First we find which session member this was.
      Person* lost = session->FindMember(packet->systemAddress);
      if(lost)
      {
        // Check if this peer triggers a reconnect - true if we were directly connected to this peer.
        bool reconnect = log->GetOwner()->ReconnectTo(lost);
        if(log->GetOwner()->DeleteConnected(lost) && reconnect)
        {
          // Set up high priority reconnect order.
          reconnectOrder.clear();
          while(reconnectOrder.size() < lost->GetConnectedCount())
          {
            // For each peer that the lost node was connected to (that we know of), add to our reconnect queue.
            size_t index = 0;
            short last = 0;
            for(size_t i=0; i<lost->GetConnectedCount(); i++)
            {
              Person* p = lost->GetConnected()[i];
              if(p->GetID() > last)
              {
                last = p->GetID();
                index = i;
              }
            }
            reconnectOrder.push_back(lost->GetConnected()[index]);
            lost->DeleteConnected(lost->GetConnected()[index]);
          }

          // Set up low priority reconnect order.
          for(size_t i=0; i<session->GetMemberCount(); i++)
          {
            // For each session member that isn't already in the list and isn't the peer we disconnected from...
            bool in = false;
            for(size_t j=0; j<reconnectOrder.size(); j++)
            {
              if(session->GetMembers()[i]->GetID() == reconnectOrder[j]->GetID() || session->GetMembers()[i]->GetID() == lost->GetID())
              {
                in = true;
              }
            }

            if(!in)
            {
              // Push to the back of the queue.
              reconnectOrder.push_back(session->GetMembers()[i]);
            }
          }

          // If there's a reconnect queue...
          if(reconnectOrder.size() != 0)
          {
            // Attempt to connect to the first in the queue.
            reconnecting = true;
            char address[16];
            reconnectOrder.begin()[0]->GetAddress().ToString(false, address);
            if(!network->Connect(address, listeningPort, 0, 0))
            {
              // Happens due to *internal* network error. Handle by cancelling the reconnect.
              reconnecting = false;
              DBG(0, QString("There was a network error! Please reconnect to the session in a few minutes."));
            }
            reconnectOrder.erase(reconnectOrder.begin());
          }
        }

        // Clean up the lost member.
        session->RemoveMember(*lost);
        MSG_LOST_PEER msg(packet->systemAddress);
        for(size_t i=0; i<log->GetOwner()->GetConnectedCount(); i++)
        {
          // Relay lost peer message across the network.
          SendBitStream(msg.GetBitStream(), HIGH_PRIORITY, RELIABLE_ORDERED, 0, &log->GetOwner()->GetConnected()[i]->GetAddress());
        }
      }
      break;
    }
  case ID_PEER_AUTHENTICATE:
    {
      Authenticate(packet);
      break;
    }
  case ID_PEER_RECONNECT:
    {
      HandleReconnect(packet);
      break;
    }
  case ID_PROPOSE_PEER:
    {
      HandleProposition(packet);
      break;
    }
  case ID_NEW_PEER:
    {
      MSG_NEW_PEER msg(packet);
      // Add the new peer data to the session.
      Person* p = new Person(msg.GetID(), msg.GetUsername(), msg.GetAddress(), CalculateTimeDifference(GetLocalTime(), msg.GetTime()), msg.GetTime());
      session->AddMember(p);

      // Replace temporary id connection info with Person connection info.
      log->GetOwner()->UpdateConnected(p);

      // Relay message to all other connected peers.
      RelayPacket(packet);

      break;
    }
  case ID_LOST_PEER:
    {
      MSG_LOST_PEER msg(packet);
      // If the peer is lost, remove it from the session.
      session->RemoveMember(*session->FindMember(msg.GetAddress()));
      break;
    }
  default:
    {
      // Relay the packet to our peers first.
      RelayPacket(packet);

      // Send the packet to all callbacks.
      bool processed = false;
      for(size_t i=0; i<receivers.size(); i++)
      {
        processed |= receivers[i]->ReceivePacket(packet, type);
      }

      // Debugging print. It's expected to get some 'unknown' messages.
#ifdef _DEBUG
      if(!processed)
      {
		    printf("Recieved unknown message with identifier %u\n", packet->data[0]);
      }
#endif

      break;
    }
  }
}

void Network::RelayPacket(Packet* &packet)
{
  // For each peer we are connected to, relay the packet.
  std::vector<Person*>::const_iterator itr(log->GetOwner()->GetConnected());
  for(size_t i=0; i<log->GetOwner()->GetConnectedCount(); i++)
  {
    if(itr[i]->GetAddress() != packet->systemAddress)
    {
      network->Send((const char*)packet->data, packet->length, HIGH_PRIORITY, (PacketReliability)packet->data[1], packet->data[2], itr[i]->GetAddress(), false);
    }
  }
}

void Network::Authenticate(Packet* &packet)
{
  MSG_PEER_AUTHENTICATE msg(packet);

  if(msg.IsConnecting())
  {
    switch(msg.GetResult())
    {
    case UNAUTHORISED:
      {
		    DBG(0, QString("Incorrect Password!"));
        
        connectActive = false;
        network->Shutdown(5000);
        break;
      }
    case INCOMPATIBLE:
      {
        DBG(0, QString("Your client is incompatible with this network!"));
        connectActive = false;
        network->Shutdown(5000);
        break;
      }
    case ALREADYIN:
      {
		    DBG(0, QString("Username already in use! Please try another."));
        
        connectActive = false;
        network->Shutdown(5000);
        break;
      }
    case SUCCESSFUL:
      {
        // Mark our username as used.
        usedUsernames.push_back(username);

        // Set 'me'.
        log->SetOwner(new Person(msg.GetMyID(), username, msg.GetSystemAddress(), 0, GetLocalTime()));

        // Add the peer we connected with to the session.
        Person* newPeer = new Person(msg.GetID(), msg.GetUsername(), packet->systemAddress, CalculateTimeDifference(GetLocalTime(), msg.GetTime()), msg.GetTime());
        session->AddMember(newPeer);
        log->GetOwner()->AddConnected(newPeer);
        log->GetOwner()->AddConnected(true);

        // Push the highest known ID to equal our own.
        usedIds = msg.GetMyID();

        // Add temp ID info to the new peer. Will be replaced when we get the proper data.
        const std::vector<short>& ids = msg.GetConnectedIDs();
        for(size_t i=0; i<ids.size(); i++)
        {
          newPeer->AddConnected(ids[i]);
        }

        DBG(0, QString("You have been successfully connnected to the network!"));

        break;
      }
    default:
      {
        break;
      }
    }
  }
  else
  {
    LOGIN_RESULT result = SUCCESSFUL;

    // Password check.
    if(password.StrCmp(msg.GetPassword()) != 0)
    {
      result = UNAUTHORISED;
    }

    // Network version check.
    if(result == SUCCESSFUL && NETVERSION != msg.GetClientVersion())
    {
      result = INCOMPATIBLE;
    }

    if(result == SUCCESSFUL)
    {
      // Check for duplicate username.
      for(size_t i=0; i<usedUsernames.size(); i++)
      {
        if(usedUsernames[i] == msg.GetUsername())
        {
          result = ALREADYIN;
          break;
        }
      }
    }

    if(result != SUCCESSFUL)
    {
      // If the authentication failed, send the appropriate message back.
      std::vector<short> unused;
      MSG_PEER_AUTHENTICATE reply(result, 0, 0, username, packet->systemAddress, 0, unused);
      SendBitStream(reply.GetResponderBitStream(), HIGH_PRIORITY, RELIABLE, 0, &packet->systemAddress);
    }
    else if(log->GetOwner()->GetConnectedCount() > 0)
    {
      // We need to propose an ID to the network.
      // Save connection data for use later.
      ConnectingPeer newPeer;
      newPeer.id = ++usedIds;
      newPeer.username = msg.GetUsername();
      newPeer.address = packet->systemAddress;
      newPeer.repliedCount = 0;
      newPeer.acceptorCount = session->GetMemberCount();
      newPeer.time = msg.GetTime();
      connecting.push_back(newPeer);

      // Propose the new ID to the network.
      MSG_PROPOSE_PEER peerProposal(newPeer.id, newPeer.username);
      std::vector<Person*>::const_iterator connectedPeers = log->GetOwner()->GetConnected();
      for(size_t i=0; i<log->GetOwner()->GetConnectedCount(); i++)
      {
        SendBitStream(peerProposal.GetProposerBitStream(), HIGH_PRIORITY, RELIABLE, 0, &connectedPeers[i]->GetAddress());
      }
    }
    else
    {
      // If this is the first peer to connect, send a success message across the network and update our own Person data.
      std::vector<short> unused;
      log->SetOwner(new Person(0, username, msg.GetSystemAddress(), 0, GetLocalTime()));
      MSG_PEER_AUTHENTICATE reply(SUCCESSFUL, log->GetOwner()->GetID(), ++usedIds, username, packet->systemAddress, GetLocalTime(), unused);
      SendBitStream(reply.GetResponderBitStream(), HIGH_PRIORITY, RELIABLE_ORDERED, 0, &packet->systemAddress);

      // Forward all events that have already happened.
      log->ForwardEvents(this, packet->systemAddress);

      // Add the new peer to the session.
      Person* newPeer = new Person(usedIds, msg.GetUsername(), packet->systemAddress, CalculateTimeDifference(GetLocalTime(), msg.GetTime()), msg.GetTime());
      session->AddMember(newPeer);
      log->GetOwner()->AddConnected(newPeer);
      log->GetOwner()->AddConnected(false);
    }
  }
}

void Network::HandleReconnect(Packet* &packet)
{
  MSG_PEER_RECONNECT msg(packet);

  if(msg.IsReconnecting())
  {
    switch(msg.GetResult())
    {
    case UNAUTHORISED:
      {
        DBG(0, QString("A network error occured and reconnect failed."));
        reconnecting = false;
        connectActive = false;
        network->Shutdown(5000);
      }
    case SUCCESSFUL:
      {
        // Add new connected peer to our connected list.
        Person* peer = session->FindMember(msg.GetID());
        log->GetOwner()->AddConnected(peer);
        log->GetOwner()->AddConnected(true);

        // Set up the temporary IDs.
        const std::vector<short>& ids = msg.GetConnectedIDs();
        for(size_t i=0; i<ids.size(); i++)
        {
          peer->AddConnected(ids[i]);
        }
        
        // Attempt to set the real data for each temp ID.
        for(size_t i=0; i<session->GetMemberCount(); i++)
        {
          peer->UpdateConnected(session->GetMembers()[i]);
        }
      }
    }
  }
  else
  {
    // If the password is valid...
    if(password == msg.GetPassword())
    {
      // And there is a matchin username and ID in the session...
      Person* p = session->FindMember(msg.GetUsername());
      if(p && p->GetID() == msg.GetID())
      {
        // Add this peer to the connected list.
        log->GetOwner()->AddConnected(p);
        log->GetOwner()->AddConnected(false);

        // Create a list of connected peers.
        std::vector<short> ids;
        for(size_t i=0; i<log->GetOwner()->GetConnectedCount(); i++)
        {
          ids.push_back(log->GetOwner()->GetConnected()[i]->GetID());
        }

        // Send a success reconnect message.
        MSG_PEER_RECONNECT reply(SUCCESSFUL, log->GetOwner()->GetID(), ids);
        SendBitStream(reply.GetResponderBitStream(), HIGH_PRIORITY, RELIABLE_ORDERED, 0, &packet->systemAddress);

        // Forward all events that have occured since the disconnect.
        log->ForwardEvents(this, packet->systemAddress, msg.GetTimestamp());

        return;
      }
    }

    // Else we send a rejection message.
    std::vector<short> unused;
    MSG_PEER_RECONNECT reply(UNAUTHORISED, log->GetOwner()->GetID(), unused);
    SendBitStream(reply.GetResponderBitStream(), HIGH_PRIORITY, RELIABLE_ORDERED, 0, &packet->systemAddress);
  }
}

void Network::HandleProposition(Packet* &packet)
{
  MSG_PROPOSE_PEER msg(packet);
  if(msg.IsFromProposer())
  {
    // Relay message to all other connected peers.
    RelayPacket(packet);

    // Check for duplicate username.
    for(size_t i=0; i<usedUsernames.size(); i++)
    {
      if(usedUsernames[i] == msg.GetUsername())
      {
        MSG_PROPOSE_PEER reply(msg.GetID(), log->GetOwner()->GetID(), BAD_USERNAME);
        SendBitStream(reply.GetAcceptorBitStream(), HIGH_PRIORITY, RELIABLE, 0, &packet->systemAddress);
        return;
      }
    }

    // Check for duplicate ID.
    if(usedIds < msg.GetID())
    {
      usedIds = msg.GetID();

      // Accept the username and ID.
      usedUsernames.push_back(msg.GetUsername());
      MSG_PROPOSE_PEER reply(msg.GetID(), log->GetOwner()->GetID(), ACCEPTED);
      SendBitStream(reply.GetAcceptorBitStream(), HIGH_PRIORITY, RELIABLE, 0, &packet->systemAddress);
    }
    else
    {
      // Check to see if it's us that's proposing.
      ConnectingPeer* newPeer = NULL;
      for(size_t i=0; i<connecting.size(); i++)
      {
        if(connecting[i].id == msg.GetID())
        {
          newPeer = &connecting[i];
          break;
        }
      }

      if(newPeer != NULL)
      {
        // We are proposing for the same ID! Send a reply back stating so ready for conflict resolution.
        MSG_PROPOSE_PEER reply(msg.GetID(), log->GetOwner()->GetID(), ALSO_PROPOSING);
        SendBitStream(reply.GetAcceptorBitStream(), HIGH_PRIORITY, RELIABLE, 0, &packet->systemAddress);
      }
      else
      {
        // It isn't, so accept the ID.
        MSG_PROPOSE_PEER reply(msg.GetID(), log->GetOwner()->GetID(), ACCEPTED);
        SendBitStream(reply.GetAcceptorBitStream(), HIGH_PRIORITY, RELIABLE, 0, &packet->systemAddress);
      }
    }
  }
  else
  {
    // Check if we're a proposer for this ID.
    size_t connectIndex;
    ConnectingPeer* newPeer = NULL;
    for(connectIndex=0; connectIndex<connecting.size(); connectIndex++)
    {
      if(connecting[connectIndex].id == msg.GetID())
      {
        newPeer = &connecting[connectIndex];
        break;
      }
    }

    if(newPeer == NULL)
    {
      // We aren't, so relay message to all other connected peers.
      RelayPacket(packet);
      return;
    }

    // Register the response.
    ++(newPeer->repliedCount);

    switch(msg.GetAccepted())
    {
    case BAD_USERNAME:
      {
        // The username was in use, tell the client that this is the case.
        std::vector<short> unused;
        MSG_PEER_AUTHENTICATE reply(ALREADYIN, log->GetOwner()->GetID(), newPeer->id, username, newPeer->address, 0, unused);
        SendBitStream(reply.GetResponderBitStream(), HIGH_PRIORITY, RELIABLE, 0, &newPeer->address);
        return;
      }
    case ALSO_PROPOSING:
      {
        // Push proposer to a list for conflict resolution later.
        newPeer->alsoProposing.push_back(msg.GetProposerID());
        break;
      }
    case ACCEPTED:
      {
        // Do nothing because accepts don't really matter, we're looking for rejections.
        break;
      }
    }

    // If all have replied then process the conflicts to see what we do.
    if(newPeer->repliedCount == newPeer->acceptorCount)
    {
      // Check for any conflicts.
      for(size_t i=0; i<newPeer->alsoProposing.size(); i++)
      {
        // If we have a higher ID than another proposer then defer.
        if(log->GetOwner()->GetID() < newPeer->alsoProposing[i])
        {
          // Try another id.
          newPeer->id = ++usedIds;

          MSG_PROPOSE_PEER peerProposal(newPeer->id, newPeer->username);
          std::vector<Person*>::const_iterator connectedPeers = log->GetOwner()->GetConnected();
          for(size_t i=0; i<log->GetOwner()->GetConnectedCount(); i++)
          {
            SendBitStream(peerProposal.GetProposerBitStream(), HIGH_PRIORITY, RELIABLE, 0, &connectedPeers[i]->GetAddress());
          }

          return;
        }
      }

      // We've won the conflict resolution by having the lowest ID! Finish the authentication process!
      std::vector<short> ids;
      for(size_t i=0; i<log->GetOwner()->GetConnectedCount(); i++)
      {
        ids.push_back(log->GetOwner()->GetConnected()[i]->GetID());
      }

      // Let the client know that the authentication was successful!
      MSG_PEER_AUTHENTICATE reply(SUCCESSFUL, log->GetOwner()->GetID(), newPeer->id, username, newPeer->address, GetLocalTime(), ids);
      SendBitStream(reply.GetResponderBitStream(), HIGH_PRIORITY, RELIABLE_ORDERED, 0, &newPeer->address);

      // Let all connected peers know about this new peer.
      MSG_NEW_PEER newPeerMsg(newPeer->id, newPeer->address, newPeer->username, newPeer->time);
      std::vector<Person*>::const_iterator connectedPeers = log->GetOwner()->GetConnected();
      for(size_t i=0; i<log->GetOwner()->GetConnectedCount(); i++)
      {
        SendBitStream(newPeerMsg.GetBitStream(), HIGH_PRIORITY, RELIABLE, 0, &connectedPeers[i]->GetAddress());
      }

      // Let the new peer know about all peers in the session.
      std::vector<Person*>::const_iterator itr(session->GetMembers());
      for(size_t i=0; i<session->GetMemberCount(); i++)
      {
        MSG_NEW_PEER oldPeer(itr[i]->GetID(), itr[i]->GetAddress(), itr[i]->GetUsername(), itr[i]->GetTime());
        SendBitStream(oldPeer.GetBitStream(), HIGH_PRIORITY, RELIABLE, 0, &newPeer->address);
      }

      // Forward all events that have occured to the new peer.
      log->ForwardEvents(this, newPeer->address);

      // Added new peer to the session and to our connected list.
      Person* newP = new Person(newPeer->id, newPeer->username, newPeer->address, CalculateTimeDifference(GetLocalTime(), newPeer->time), newPeer->time);
      session->AddMember(newP);
      log->GetOwner()->AddConnected(newP);
      log->GetOwner()->AddConnected(false);

      // Remove from the vector of 'connecting' peers.
      connecting.erase(connecting.begin() + connectIndex);
    }
  }
}

short Network::CalculateTimeDifference(short localTime, short foreignTime)
{
  if(localTime == foreignTime)
  {
    return 0;
  }

  short timeDifference;

  if(localTime > foreignTime)
  {
    timeDifference = -1 * (localTime - foreignTime);
  }
  else
  {
    timeDifference = foreignTime - localTime;
  }

  if(timeDifference > 720)
  {
    timeDifference -= 720;
  }

  return timeDifference;
}

short Network::GetLocalTime()
{
  short time = QTime::currentTime().hour()*60;
  short minutes = QTime::currentTime().minute();
  if(minutes > 30 && minutes < 45 || minutes < 30 && minutes > 15)
  {
    time += 30;
  }
  
  return time;
}

void Network::SendBitStream(RakNet::BitStream& bs, PacketPriority priority, PacketReliability reliability, char orderingChannel, const SystemAddress* address)
{
  if(!reconnecting)
    network->Send(&bs, priority, reliability, 0, (address != 0 ? *address : SystemAddress()), (address == 0));
}
