/*
* network.h
*
* Copyright (C) 2008 Mike Gist, Jan Hosang, Daniel Burrell, Dave Lawrence, Andrew Slater
*
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation (version 2 of the License)
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*/

#ifndef __NETWORK_H__
#define __NETWORK_H__

//====================================================================================
// QT Includes
//====================================================================================
#include <QtCore/QObject>

//====================================================================================
// Project Includes
//====================================================================================
#include "inetwork.h"
#include "ipacketreceiver.h"
#include "person.h"
#include "raknet/RakPeerInterface.h"
#include "raknet/RakString.h"
#include <vector>

//====================================================================================
// Local Includes
//====================================================================================

//------------------------------------------------------------------------------------
// Forward Declarations and Preprocessor Definitions
//------------------------------------------------------------------------------------
class Chat;
struct iLog;
struct iSession;

class Network : public iNetwork, public QObject
{
public:
  Network();
  virtual ~Network();

  static Network& GetSingleton()
  {
    if(!myself)
    {
      myself = new Network();
    }

    return *myself;
  }

  static Network* GetSingletonPtr()
  {
    return &GetSingleton();
  }

  /// Start a server or connect to a tracker.
  void Connect(int listeningPort, int connectingPort, const char* ip, const char* username,
    const char* password, bool connecting = false);

  /// Disconnect and clean up.
  static void Disconnect()
  {
    if(myself)
      delete myself;
  }

  /// Broadcast a bitstream.
  virtual void SendBitStream(RakNet::BitStream& bs, PacketPriority priority, PacketReliability reliability, char orderingChannel, const SystemAddress* address);

  inline void RegisterPacketReciever(iPacketReceiver* iPR)
  {
    receivers.push_back(iPR);
  }

  inline void SetSession(iSession* session)
  {
    this->session = session;
  }

  inline void SetLog(iLog* log)
  {
    this->log = log;
  }

public slots:
  void Process();

private:
  /// Authentication
  void Authenticate(Packet* &packet);

  /// Singleton object pointer.
  static Network* myself;

  /// Process a single packet.
  void ProcessPacket(Packet* &packet, unsigned char type);

  /// Handle a new peer proposition.
  void HandleProposition(Packet* &packet);

  /// Handle a peer reconnect.
  void HandleReconnect(Packet* &packet);

  /// Relay a packet to all connected peers.
  void RelayPacket(Packet* &packet);

  /// Calculate the time difference between two timezones.
  short CalculateTimeDifference(short localTime, short foreignTime);

  /// Get the local time correct to the nearest half an hour.
  short GetLocalTime();

  /// True when we're connected or connecting to the tracker.
	bool connectActive;

  /// Current session.
  iSession* session;

  /// Pointer to the log.
  iLog* log;

  /// Pointer to main raknet object.
  RakPeerInterface* network;

  /// The ip we can't tried to connect to.
  RakNet::RakString ip;

  /// The port that we're listening on.
  int listeningPort;

  /// Our username.
  RakNet::RakString username;

  /// Session password.
  RakNet::RakString password;

  /// Registered packet recievers.
  std::vector<iPacketReceiver*> receivers;

  /// Used id's.
  short usedIds;

  /// Used usernames.
  std::vector<RakNet::RakString> usedUsernames;

  /// struct of connecting peer data.
  struct ConnectingPeer
  {
    short id;
    SystemAddress address;
    RakNet::RakString username;
    unsigned short repliedCount;
    std::vector<short> alsoProposing;
    size_t acceptorCount;
    short time;
  };

  /// Currently connecting peers.
  std::vector<ConnectingPeer> connecting;

  /// Reconnect order.
  std::vector<const Person*> reconnectOrder;

  /// True if we're attempting a reconnect.
  bool reconnecting;

Q_OBJECT
};

#endif // __NETWORK_H__
